import os
import logging
import logging.config
from pathlib import Path

from bunch import Bunch
from azureml.core import Experiment, Workspace
from azureml.core.run import Run

from cgmml.common.model_utils.utils import download_dataset, get_dataset_path

DATA_DIR_ONLINE_RUN = Path("/tmp/data/")
REPO_DIR = Path(__file__).parents[3].absolute()

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


class RunInitializer:
    """Setup AzureML and prepare dataset"""
    def __init__(self, data_config: Bunch, eval_config: Bunch):
        self._data_config = data_config
        self._eval_config = eval_config
        self.run_azureml_setup()
        self.get_dataset()

    def run_azureml_setup(self):
        raise NotImplementedError

    def get_dataset(self):
        raise NotImplementedError


class OfflineRunInitializer(RunInitializer):
    """Offline run. Download the sample dataset and run locally. Still push results to Azure"""
    def __init__(self, data_config: Bunch, eval_config: Bunch):
        super().__init__(data_config, eval_config)

    def run_azureml_setup(self):
        logger.info("Running in offline mode...")
        logger.info("Accessing workspace...")
        self.workspace = Workspace.from_config()
        self.experiment = Experiment(self.workspace, self._eval_config.EXPERIMENT_NAME)
        self.run = self.experiment.start_logging(outputs=None, snapshot_directory=None)

    def get_dataset(self):
        logger.info("Accessing dataset...")
        dataset_name = self._data_config.NAME
        self.dataset_path = str(REPO_DIR / "data" / dataset_name)
        if not os.path.exists(self.dataset_path):
            dataset = self.workspace.datasets[dataset_name]
            dataset.download(target_path=self.dataset_path, overwrite=False)


class OnlineRunInitializer(RunInitializer):
    def __init__(self, data_config: Bunch, eval_config: Bunch, run: Run):
        self.run = run
        super().__init__(data_config, eval_config)

    def run_azureml_setup(self):
        logger.info("Running in online mode...")
        self.experiment = self.run.experiment
        self.workspace = self.experiment.workspace

    def get_dataset(self):
        dataset_name = self._data_config.NAME
        # Download
        self.dataset_path = get_dataset_path(DATA_DIR_ONLINE_RUN, dataset_name)
        download_dataset(self.workspace, dataset_name, self.dataset_path)



import sys
from csv_utils import read_csv, write_csv

METADATA_SCAN_ID = 0
METADATA_ORDER = 1
METADATA_ARTIFACT_ID = 2
METADATA_FILEPATH = 3
METADATA_EXTENSION = 4


if __name__ == "__main__":

    if len(sys.argv) != 2:
        print('You did not enter metadata file path')
        print('E.g.: python rgbd_match.py metadata_path')
        sys.exit(1)

    metadata_file = sys.argv[1]

    # Create a map
    indata = read_csv(metadata_file)
    size = len(indata)
    mapping = {}
    for index in range(1, size):
        data = indata[index]
        if data[METADATA_EXTENSION] != '.jpg':
            continue
        scanid = data[METADATA_SCAN_ID]
        order = data[METADATA_ORDER]
        key = scanid + str(order)
        mapping[key] = data

    # For every depthmap add rgb file
    output = []
    processed = {}
    for index in range(1, size):
        data = indata[index]
        if data[METADATA_EXTENSION] != '.depth':
            continue
        scanid = data[METADATA_SCAN_ID]
        order = data[METADATA_ORDER]
        key = scanid + str(order)
        if processed.get(key):
            continue

        # add the RGB-D match
        if mapping.get(key):
            data[METADATA_EXTENSION] = mapping[key][METADATA_FILEPATH]
            print(str(index) + '/' + str(size) + ' - ' + scanid + ',' + str(order))
            output.append(data)
            processed[key] = True
    write_csv('newmetadata.csv', output)



import json
from decimal import Decimal as D

from .calculate import Zscore
from .constant import REPO_DIR, WHO_TABLES
from .extract_value import LmsBoxCox
from .table import Table
TABLE_REPO = REPO_DIR / 'tables'
GROWTH_CHART = {}


def setup_tables():
    for t in WHO_TABLES:
        table_name, _, _ = t.split('.')[0].rpartition('_')
        table = Table(TABLE_REPO / t)
        table.load_table()
        table.add_value()
        new_table = table.append_value()
        GROWTH_CHART[table_name] = new_table


def z_score_calculation(lms_box_cox: LmsBoxCox) -> float:
    lms_box_cox.get_file_name_from_data()
    lms_box_cox.get_lms_value(GROWTH_CHART)
    skew, median, coff, measurement = lms_box_cox.resolve_lms_value()
    z_score_value = Zscore(skew, median, coff, measurement).z_score_measurement()
    return z_score_value


def z_score_wfa(weight: str, age_in_days: str, sex: str) -> int:
    """z-score for weight for age"""
    setup_tables()

    lms_box_cox = LmsBoxCox('wfa', weight=weight, muac=None, age_in_days=age_in_days, sex=sex, height=None)
    return z_score_calculation(lms_box_cox)


def z_score_wfh(weight: str, age_in_days: str, sex: str, height: str) -> int:
    """z-score for weight for height"""
    setup_tables()

    if D(age_in_days) <= 731:
        return z_score_wfl(weight, age_in_days, sex, height)
    lms_box_cox = LmsBoxCox('wfh', weight=weight, muac=None, age_in_days=age_in_days, sex=sex, height=height)
    return z_score_calculation(lms_box_cox)


def z_score_wfl(weight: str, age_in_days: str, sex: str, height: str) -> int:
    """z-score for weight for length"""
    setup_tables()
    if D(age_in_days) > 731:
        return z_score_wfh(weight, age_in_days, sex, height)
    lms_box_cox = LmsBoxCox('wfl', weight=weight, muac=None, age_in_days=age_in_days, sex=sex, height=height)
    return z_score_calculation(lms_box_cox)


def z_score_lhfa(age_in_days: str, sex: str, height: str) -> int:
    """z-score for length/height for age"""
    setup_tables()
    lms_box_cox = LmsBoxCox('lhfa', weight=None, muac=None, age_in_days=age_in_days, sex=sex, height=height)
    return z_score_calculation(lms_box_cox)


def z_score_with_class(weight: str, muac: str, age_in_days: str, sex: str, height: str):
    wfa = z_score_wfa(weight=weight, age_in_days=age_in_days, sex=sex)
    if wfa < -3:
        class_wfa = 'Severely Under-weight'
    elif -3 <= wfa < -2:
        class_wfa = 'Moderately Under-weight'
    else:
        class_wfa = 'Healthy'

    wfl = z_score_wfl(weight=weight, age_in_days=age_in_days, sex=sex, height=height)
    class_wfl = calculate_sam_mam(weight, muac, age_in_days, sex, height)

    lhfa = z_score_lhfa(age_in_days=age_in_days, sex=sex, height=height)
    if lhfa < -3:
        class_lhfa = 'Severely Stunted'
    elif -3 <= lhfa < -2:
        class_lhfa = 'Moderately Stunted'
    else:
        class_lhfa = 'Healthy'

    return json.dumps({'Z_score_WFA': wfa, 'Class_WFA': class_wfa, 'Z_score_WFH' if D(age_in_days) > 24 else
                       'Z_score_WFL': wfl, 'Class_WFH' if D(age_in_days) > 24 else 'Class_WFL': class_wfl,
                       'Z_score_HFA': lhfa, 'Class_HFA': class_lhfa})


def calculate_sam_mam(weight: str, muac: str, age_in_days: str, sex: str, height: str) -> str:
    assert muac is not None

    if D(age_in_days) > 731:
        wfl = z_score_wfh(
            weight=weight, age_in_days=age_in_days, sex=sex, height=height)
    else:
        wfl = z_score_wfl(
            weight=weight, age_in_days=age_in_days, sex=sex, height=height)
    if wfl < -3 or D(muac) < 11.5:
        diagnosis = "SAM"
    elif -3 <= wfl < -2 or D(muac) < 12.5:
        diagnosis = "MAM"
    else:
        diagnosis = "Healthy"
    return diagnosis



import os
import logging
import logging.config
from pathlib import Path
import pickle
import tensorflow as tf

from bunch import Bunch
import pandas as pd

from .constants_eval import (  # noqa: E402, F401
    AGE_IDX, COLUMN_NAME_AGE, COLUMN_NAME_GOODBAD, COLUMN_NAME_SEX, CONFIG, EVALUATION_ACCURACIES,
    GOODBAD_DICT, GOODBAD_IDX, HEIGHT_IDX, SEX_DICT, SEX_IDX, WEIGHT_IDX)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

REPO_DIR = Path(os.getcwd()).parents[2]


def calculate_performance_mae(code: str,
                              df_mae: pd.DataFrame,
                              result_config: Bunch) -> pd.DataFrame:
    df_mae_filtered = df_mae.iloc[df_mae.index.get_level_values('scantype') == code]
    mae = df_mae_filtered['error'].abs().mean()
    df_out = pd.DataFrame.from_dict({'mae': [mae]})
    return df_out


def calculate_performance(code: str,
                          df_mae: pd.DataFrame,
                          result_config: Bunch) -> pd.DataFrame:
    """For a specific scantype, calculate the performance of the model on each error margin
    Args:
        code: e.g. '100'
        df_mae: dataframe
        result_config: bunch containing result config
    Returns:
        dataframe, where each column describes a different accuracy, e.g.
                            0.2   0.4   0.6   1.0   1.2    2.0    2.5    3.0    4.0    5.0    6.0
                           20.0  20.0  40.0  80.0  80.0  100.0  100.0  100.0  100.0  100.0  100.0
    """
    df_mae_filtered = df_mae.iloc[df_mae.index.get_level_values('scantype') == code]
    accuracy_list = []
    for acc in result_config.ACCURACIES:
        good_predictions = df_mae_filtered[(df_mae_filtered['error'] <= acc) & (df_mae_filtered['error'] >= -acc)]
        if len(df_mae_filtered) > 0:
            accuracy = len(good_predictions) / len(df_mae_filtered) * 100
        else:
            accuracy = 0.
        accuracy_list.append(accuracy)
    df_out = pd.DataFrame(accuracy_list)
    df_out = df_out.T
    df_out.columns = result_config.ACCURACIES
    return df_out


# Function for loading and processing depthmaps.
def tf_load_pickle(path, max_value):
    def py_load_pickle(path, max_value):
        depthmap, targets = pickle.load(open(path.numpy(), "rb"))
        depthmap = preprocess_depthmap(depthmap)
        depthmap = depthmap / max_value
        depthmap = tf.image.resize(depthmap, (CONFIG.IMAGE_TARGET_HEIGHT, CONFIG.IMAGE_TARGET_WIDTH))
        targets = preprocess_targets(targets, CONFIG.TARGET_INDEXES)
        return depthmap, targets

    depthmap, targets = tf.py_function(py_load_pickle, [path, max_value], [tf.float32, tf.float32])
    depthmap.set_shape((CONFIG.IMAGE_TARGET_HEIGHT, CONFIG.IMAGE_TARGET_WIDTH, 1))
    targets.set_shape((len(CONFIG.TARGET_INDEXES,)))
    return depthmap, targets


def extract_qrcode(row):
    qrc = row['artifacts'].split('/')[-3]
    return qrc


def extract_scantype(row):
    """https://dev.azure.com/cgmorg/ChildGrowthMonitor/_wiki/wikis/ChildGrowthMonitor.wiki/15/Codes-for-Pose-and-Scan-step"""  # noqa: E501
    scans = row['artifacts'].split('/')[-2]
    return scans


def avgerror(row):
    difference = row['GT'] - row['predicted']
    return difference


def preprocess(path):
    depthmap, targets = pickle.load(open(path, "rb"))
    depthmap = preprocess_depthmap(depthmap)
    depthmap = depthmap / CONFIG.NORMALIZATION_VALUE
    depthmap = tf.image.resize(depthmap, (CONFIG.IMAGE_TARGET_HEIGHT, CONFIG.IMAGE_TARGET_WIDTH))
    targets = preprocess_targets(targets, CONFIG.TARGET_INDEXES)
    depthmap.set_shape((CONFIG.IMAGE_TARGET_HEIGHT, CONFIG.IMAGE_TARGET_WIDTH, 1))
    return depthmap, targets


def preprocess_targets(targets, targets_indices):
    if SEX_IDX in targets_indices:
        targets[SEX_IDX] = SEX_DICT[targets[SEX_IDX]]
    if GOODBAD_IDX in targets_indices:
        try:
            targets[GOODBAD_IDX] = GOODBAD_DICT[targets[GOODBAD_IDX]]
        except KeyError:
            logger.info("Key %s not found in GOODBAD_DICT", targets[GOODBAD_IDX])
            targets[GOODBAD_IDX] = GOODBAD_DICT['delete']  # unknown target values will be categorized as 'delete'

    if targets_indices is not None:
        targets = targets[targets_indices]
    return targets.astype("float32")


def preprocess_depthmap(depthmap):
    return depthmap.astype("float32")



from collections import namedtuple
import zipfile
import math
import sys
from typing import List, Optional, Tuple
from pathlib import Path

from scipy import ndimage
import numpy as np
from PIL import Image

from cgmml.common.depthmap_toolkit.depthmap_utils import (
    matrix_calculate, IDENTITY_MATRIX_4D, parse_numbers, calculate_boundary)
from cgmml.common.depthmap_toolkit.constants import EXTRACTED_DEPTH_FILE_NAME, MASK_FLOOR, MASK_CHILD, MASK_INVALID

NEIGHBOUR_PIXELS_MAX_DISTANCE_IN_METER = 0.05
FLOOR_THRESHOLD_IN_METER = 0.1
SCREEN_BORDER_MARGIN_PERCENTAGE = 0.05
SCREEN_BORDER_PERCENTAGE = 0.025
STANDING_CHILD_MAX_HORIZONTAL_DIFF_IN_METER = 0.075
TOOLKIT_DIR = Path(__file__).parents[0].absolute()


Segment = namedtuple('Segment', 'id aabb')


def extract_depthmap(depthmap_fpath: str) -> Path:
    """Extract depthmap from given file"""
    with zipfile.ZipFile(Path(depthmap_fpath), 'r') as zip_ref:
        zip_ref.extractall(TOOLKIT_DIR)
    return TOOLKIT_DIR / EXTRACTED_DEPTH_FILE_NAME


def smoothen_depthmap_array(image_arr: np.ndarray) -> np.ndarray:
    """Smoothen image array by averaging with direct neighbor pixels.

    Args:
        image_arr: shape (width, height)

    Returns:
        shape (width, height)
    """

    # Apply a convolution
    conv_filter = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]])
    conv_filter = conv_filter / conv_filter.sum()
    smooth_image_arr = ndimage.convolve(image_arr, conv_filter)

    # Create mask which is 0 if any of the pixels in the convolution is 0
    smooth_center = (image_arr == 0.)
    smooth_right = np.zeros(image_arr.shape, dtype=bool)
    smooth_left = np.zeros(image_arr.shape, dtype=bool)
    smooth_up = np.zeros(image_arr.shape, dtype=bool)
    smooth_down = np.zeros(image_arr.shape, dtype=bool)
    smooth_right[1:, :] = smooth_center[:-1, :]
    smooth_left[:-1, :] = smooth_center[1:, :]
    smooth_up[:, 1:] = smooth_center[:, :-1]
    smooth_down[:, :-1] = smooth_center[:, 1:]
    mask = (smooth_center | smooth_right | smooth_left | smooth_up | smooth_down)

    # Apply mask
    smooth_image_arr[mask] = 0.

    return smooth_image_arr


class Depthmap:
    """Depthmap and optional RGB

    Args:
        intrinsic (np.array): Camera intrinsic
        width (int): The longer side of the depthmap (and RGB
        height (int): The shorter side of the depthmap (and RGB
        data (bytes): pixel_data
        depth_scale (float): Scalar to scale depthmap pixel to meters
        max_confidence (float): Confidence is amount of IR light reflected
                                (e.g. 0 to 255 in Lenovo, new standard is 0 to 7)
                                This is actually an int.
        device_pose (List[float]): The device pose (= position and rotation)
                              The ZIP-file header contains this pose
                              - `device_pose` is a list representation of this pose
                              - can be used to project into a different space
        rgb_fpath (str): Path to RGB file (e.g. to the jpg)
        rgb_array (np.array): RGB data
        header: raw depthmap header with metainformation
    """
    def __init__(
            self,
            intrinsics: np.ndarray,
            width: int,
            height: int,
            data: Optional[bytes],
            depthmap_arr: Optional[np.array],
            depth_scale: float,
            max_confidence: float,
            device_pose: List[float],
            rgb_fpath: Path,
            rgb_array: np.ndarray,
            header: str = None):
        """Constructor

        Either `data` or `depthmap_arr` has to be defined
        """
        self.width = width
        self.height = height

        self.intrinsics = np.array(intrinsics)
        self.sensor = 1
        self.fx = self.intrinsics[self.sensor, 0] * self.width
        self.fy = self.intrinsics[self.sensor, 1] * self.height
        self.cx = self.intrinsics[self.sensor, 2] * self.width
        self.cy = self.intrinsics[self.sensor, 3] * self.height

        self.depth_scale = depth_scale
        self.max_confidence = max_confidence
        self.device_pose = device_pose
        self.device_pose_arr = np.array(device_pose).reshape(4, 4).T
        if header:
            self.header = header

        self.rgb_fpath = rgb_fpath
        self.rgb_array = rgb_array  # shape (width, height, 3)  # (180, 240, 3)

        assert depthmap_arr is not None or data is not None
        self.depthmap_arr = self._parse_depth_data(data) if data else depthmap_arr  # (240, 180)

        assert self.depthmap_arr.shape[:2] == (self.width, self.height)
        if self.rgb_array is not None:
            assert self.rgb_array.shape[:2] == (self.width, self.height)

        # smoothing is only for normals, otherwise there is noise
        self.depthmap_arr_smooth = smoothen_depthmap_array(self.depthmap_arr)

        self.confidence_arr = self._parse_confidence_data(data) if data else None

    @property
    def has_rgb(self) -> bool:
        """Bool that indicates if the object has RGB data"""
        return self.rgb_array is not None

    @classmethod
    def create_from_zip_absolute(cls,
                                 depthmap_fpath: str,
                                 rgb_fpath: str,
                                 calibration_fpath: str,
                                 should_rotate_rgb: bool = True) -> 'Depthmap':
        # read depthmap data
        path = extract_depthmap(depthmap_fpath)
        with open(path, 'rb') as f:
            header_line = f.readline().decode().strip()
            header_parts = header_line.split('_')
            res = header_parts[0].split('x')
            width = int(res[0])
            height = int(res[1])
            depth_scale = float(header_parts[1])
            max_confidence = float(header_parts[2])
            if len(header_parts) >= 10:
                position = (float(header_parts[7]), float(header_parts[8]), float(header_parts[9]))
                rotation = (float(header_parts[3]), float(header_parts[4]),
                            float(header_parts[5]), float(header_parts[6]))
                device_pose = matrix_calculate(position, rotation)
            else:
                device_pose = IDENTITY_MATRIX_4D
            data = f.read()
            f.close()

        # read rgb data
        if rgb_fpath:
            pil_im = Image.open(rgb_fpath)
            if should_rotate_rgb:
                pil_im = pil_im.rotate(-90, expand=True)
            rgb_height, rgb_width = pil_im.width, pil_im.height  # Weird switch
            assert rgb_width / width == rgb_height / height, f'{rgb_width} / {width} != {rgb_height} / {height}'
            pil_im = pil_im.resize((height, width), Image.ANTIALIAS)
            rgb_array = np.asarray(pil_im)
        else:
            rgb_array = None

        # read calibration file
        intrinsics = parse_calibration(calibration_fpath)
        depthmap_arr = None

        return cls(intrinsics,
                   width,
                   height,
                   data,
                   depthmap_arr,
                   depth_scale,
                   max_confidence,
                   device_pose,
                   rgb_fpath,
                   rgb_array,
                   header_line,
                   )

    @classmethod
    def create_from_zip(cls,
                        depthmap_dir: str,
                        depthmap_fname: str,
                        rgb_fname: str,
                        calibration_file: str,
                        should_rotate_rgb: bool = False) -> 'Depthmap':

        depthmap_path = Path(depthmap_dir) / 'depth' / depthmap_fname
        rgb_fpath = 0
        if rgb_fname:
            rgb_fpath = Path(depthmap_dir) / 'rgb' / rgb_fname
        return cls.create_from_zip_absolute(depthmap_path, rgb_fpath, calibration_file, should_rotate_rgb)

    @classmethod
    def create_from_array(cls,
                          depthmap_arr: np.ndarray,
                          rgb_arr: np.ndarray,
                          calibration_file: str) -> 'Depthmap':
        intrinsics = parse_calibration(calibration_file)
        height, width = depthmap_arr.shape
        data = None  # bytes
        depth_scale = 0.001
        max_confidence = 7.0
        device_pose = None
        rgb_fpath = None
        rgb_array = rgb_arr

        return cls(intrinsics,
                   width,
                   height,
                   data,
                   depthmap_arr,
                   depth_scale,
                   max_confidence,
                   device_pose,
                   rgb_fpath,
                   rgb_array,
                   )

    def calculate_normalmap_array(self, points_3d_arr: np.ndarray) -> np.ndarray:
        """Calculate normalmap consisting of normal vectors.

        A normal vector is based on a surface.
        The surface is constructed by a 3D point and it's neighbors.

        points_3d_arr: shape (3, width, height)

        Returns:
            3D points: shape (3, width, height)
        """

        # Get depth of the neighbor pixels
        dim_w = self.width - 1
        dim_h = self.height - 1
        depth_center = points_3d_arr[:, 1:, 1:].reshape(3, dim_w * dim_h)
        depth_x_minus = points_3d_arr[:, 0:-1, 1:].reshape(3, dim_w * dim_h)
        depth_y_minus = points_3d_arr[:, 1:, 0:-1].reshape(3, dim_w * dim_h)

        # Calculate a normal of the triangle
        vector_u = depth_center - depth_x_minus
        vector_v = depth_center - depth_y_minus

        normal = np.cross(vector_u, vector_v, axisa=0, axisb=0, axisc=0)

        normal = normalize(normal)

        normal = normal.reshape(3, dim_w, dim_h)

        # add black border to keep the dimensionality
        output = np.zeros((3, self.width, self.height))
        output[:, 1:, 1:] = normal
        return output

    def calculate_standing_confidence(self, mask: np.array, highest_point: list) -> float:
        """Calculates the confidence if the child is standing straight.

        The result is calculated by calculating difference of x,z coordinates of highest point and every child point.
        The more unstraight child stays, the bigger the difference is.
        The confidence is 1.0 when the child is fully straight, 0.5 when it could be still count as straight.
        """

        points_3d_arr = self.convert_2d_to_3d_oriented(should_smooth=False)

        # count average horizontal diff of child points to highest_point
        count = 0
        horizontal_diff = 0.
        for x in range(self.width):
            for y in range(self.height):
                if mask[x, y] != MASK_CHILD:
                    continue

                diff_x = points_3d_arr[0, x, y] - highest_point[0]
                diff_z = points_3d_arr[2, x, y] - highest_point[2]
                horizontal_diff += math.sqrt(diff_x * diff_x + diff_z * diff_z)
                count += 1
        if count > 0:
            horizontal_diff /= float(count)

        # STANDING_CHILD_MAX_HORIZONTAL_DIFF_IN_METER for confidence=0.5, fully straight standing for confidence=1
        confidence = 1.0 - horizontal_diff / STANDING_CHILD_MAX_HORIZONTAL_DIFF_IN_METER / 2.0
        return min(confidence, 1.)

    def convert_2d_to_3d(self, sensor: int, x: float, y: float, depth: float) -> np.ndarray:
        """Convert point in pixels into point in meters

        Args:
            sensor: Index of sensor
            x
            y
            depth

        Returns:
            3D point
        """
        fx = self.intrinsics[sensor, 0] * self.width
        fy = self.intrinsics[sensor, 1] * self.height
        cx = self.intrinsics[sensor, 2] * self.width
        cy = self.intrinsics[sensor, 3] * self.height
        tx = (x - cx) * depth / fx
        ty = (y - cy) * depth / fy
        return np.array([tx, ty, depth])

    def convert_2d_to_3d_oriented(self, should_smooth: bool = False) -> np.ndarray:
        """Convert points in pixels into points in meters (and applying rotation)

        Args:
            should_smooth: Flag indicating weather to use a smoothed or an un-smoothed depthmap

        Returns:
            array of 3D points: shape(3, width, height)
        """
        depth = self.depthmap_arr_smooth if should_smooth else self.depthmap_arr  # shape: (width, height)

        xbig = np.expand_dims(np.array(range(self.width)), -1).repeat(self.height, axis=1)  # shape: (width, height)
        ybig = np.expand_dims(np.array(range(self.height)), 0).repeat(self.width, axis=0)  # shape: (width, height)

        # Convert point in pixels into point in meters
        tx = depth * (xbig - self.cx) / self.fx
        ty = depth * (ybig - self.cy) / self.fy
        dim4 = np.ones((self.width, self.height))
        res = np.stack([-tx, ty, depth, dim4], axis=0)

        # Transformation of point by device pose matrix
        points_4d = res.reshape((4, self.width * self.height))
        output = np.matmul(self.device_pose_arr, points_4d)
        output[0:2, :] = output[0:2, :] / abs(output[3, :])
        output = output.reshape((4, self.width, self.height))
        res = output[0:-1]

        # Invert y axis
        res[1, :, :] = -res[1, :, :]
        return res

    def is_child_fully_visible(self, mask: np.array) -> bool:

        # Get the boundary of child and of valid data
        margin = max(self.width, self.height) * SCREEN_BORDER_MARGIN_PERCENTAGE
        child_aabb = calculate_boundary(mask == MASK_CHILD)
        valid_aabb = calculate_boundary(self.depthmap_arr > 0)

        # Check if the child boundary is inside valid boundary
        if child_aabb[0] < valid_aabb[0] + margin:
            return False
        if child_aabb[1] < valid_aabb[1] + margin:
            return False
        if child_aabb[2] > valid_aabb[2] - margin:
            return False
        if child_aabb[3] > valid_aabb[3] - margin:
            return False
        return True

    def segment_child(self, floor: float) -> np.ndarray:
        """Segment the child from the background

        Args:
            floor: At which y-value the floor is

        Returns:
            np.ndarray (int): Each pixel has either
                * -1, -2, -3 (according to number of objects in the scene)
                * MASK_FLOOR = 1
                * MASK_CHILD = 2
                * MASK_INVALID = 3
        """
        mask, segments = self.detect_objects(floor)

        # Select the most focused segment
        closest = sys.maxsize
        focus = -1
        for segment in segments:
            a = segment.aabb[0] - int(self.width / 2)
            b = segment.aabb[1] - int(self.height / 2)
            c = segment.aabb[2] - int(self.width / 2)
            d = segment.aabb[3] - int(self.height / 2)
            distance = a * a + b * b + c * c + d * d
            if closest > distance:
                closest = distance
                focus = segment.id

        mask = np.where(mask == focus, MASK_CHILD, mask)

        assert not (mask == 0).any()  # 0 pixels in output_mask are not allowed

        return mask

    def detect_floor(self, floor: float) -> np.ndarray:
        mask = np.zeros((self.width, self.height))
        assert self.depthmap_arr_smooth.shape == (self.width, self.height)
        mask[self.depthmap_arr_smooth == 0] = MASK_INVALID

        points_3d_arr = self.convert_2d_to_3d_oriented(should_smooth=True)
        normal = self.calculate_normalmap_array(points_3d_arr)

        cond1 = np.abs(normal[1, :, :]) > 0.5
        cond2 = (points_3d_arr[1, :, :] - floor) < FLOOR_THRESHOLD_IN_METER
        per_pixel_cond = cond1 & cond2
        mask[per_pixel_cond] = MASK_FLOOR
        return mask

    def detect_objects(self, floor: float) -> Tuple[np.array, List[Segment]]:
        """Detect objects/children using seed algorithm

        Can likely not be used without for-loops over x,y

        Args:
            floor: Value of y-coordinate where the floor is

        Returns:
            mask (np.array): binary mask
            List[Segment]: a list of segments
        """
        current_id = -1
        segments = []
        dirs = [[-1, 0], [1, 0], [0, -1], [0, 1]]
        mask = self.detect_floor(floor)
        points_3d_arr = self.convert_2d_to_3d_oriented(should_smooth=False)
        for x in range(self.width):
            for y in range(self.height):
                if mask[x, y] != 0:
                    continue
                pixel = [x, y]
                stack = [pixel]
                while len(stack) > 0:

                    # Get a next pixel from the stack
                    pixel = stack.pop()
                    depth_center = self.depthmap_arr[pixel[0], pixel[1]]

                    # Add neighbor points (if there is no floor and they are connected)
                    if mask[pixel[0], pixel[1]] == 0:
                        if points_3d_arr[1, pixel[0], pixel[1]] - floor > FLOOR_THRESHOLD_IN_METER:
                            for direction in dirs:
                                pixel_dir = [pixel[0] + direction[0], pixel[1] + direction[1]]
                                depth_dir = self.depthmap_arr[pixel_dir[0], pixel_dir[1]]
                                if depth_dir > 0 and abs(
                                        depth_dir - depth_center) < NEIGHBOUR_PIXELS_MAX_DISTANCE_IN_METER:
                                    stack.append(pixel_dir)

                    # Update the mask
                    mask[pixel[0], pixel[1]] = current_id

                # Check if the object size is valid
                aabb = calculate_boundary(mask == current_id)
                object_size_pixels = max(aabb[2] - aabb[0], aabb[3] - aabb[1])
                if object_size_pixels > self.width / 4:
                    segments.append(Segment(current_id, aabb))
                current_id = current_id - 1

        return mask, segments

    def get_angle_between_camera_and_floor(self) -> float:
        """Calculate an angle between camera and floor based on device pose"""
        centerx = int(self.width / 2)
        centery = int(self.height / 2)
        points_3d_arr = self.convert_2d_to_3d_oriented()  # shape: (3, width, height)
        # TODO revert to original code (depth=1 is important)

        point = points_3d_arr[:, centerx, centery]  # shape: (3,)
        angle = 90 + math.degrees(math.atan2(point[0], point[1]))
        return angle

    def get_floor_level(self) -> float:
        """Calculate an altitude of the floor in the world coordinates"""

        # Get normal vectors
        mask = np.zeros((self.width, self.height))
        assert self.depthmap_arr_smooth.shape == (self.width, self.height)
        mask[self.depthmap_arr_smooth == 0] = MASK_INVALID
        points_3d_arr = self.convert_2d_to_3d_oriented(should_smooth=True)
        normal = self.calculate_normalmap_array(points_3d_arr)

        cond = np.abs(normal[1, :, :]) > 0.5
        selection_of_points = points_3d_arr[1, :, :][cond]
        median = np.median(selection_of_points)
        return median

    def get_highest_point(self, mask: np.ndarray) -> np.ndarray:
        points_3d_arr = self.convert_2d_to_3d_oriented()
        y_array = np.copy(points_3d_arr[1, :, :])
        y_array[mask != MASK_CHILD] = -np.inf
        idx_highest_child_point = np.unravel_index(np.argmax(y_array, axis=None), y_array.shape)
        highest_point = points_3d_arr[:, idx_highest_child_point[0], idx_highest_child_point[1]]
        return highest_point

    def _parse_confidence_data(self, data) -> np.ndarray:
        """Parse depthmap confidence

        Returns:
            2D array of floats
        """
        output = np.zeros((self.width, self.height))
        for x in range(self.width):
            for y in range(self.height):
                output[x, y] = self._parse_confidence(data, x, y)
        return output

    def _parse_confidence(self, data: bytes, tx: int, ty) -> float:
        """Get confidence of the point in scale 0-1"""
        return data[(int(ty) * self.width + int(tx)) * 3 + 2] / self.max_confidence

    def _parse_depth_data(self, data) -> np.ndarray:
        output = np.zeros((self.width, self.height))
        for x in range(self.width):
            for y in range(self.height):
                output[x, y] = self._parse_depth(data, x, y)
        return output

    def _parse_depth(self, data: bytes, tx: int, ty: int) -> float:
        """Get depth of the point in meters"""
        if tx < 1 or ty < 1 or tx >= self.width or ty >= self.height:
            return 0.
        depth = data[(int(ty) * self.width + int(tx)) * 3 + 0] << 8
        depth += data[(int(ty) * self.width + int(tx)) * 3 + 1]
        depth *= self.depth_scale
        return depth


def convert_3d_to_2d(intrinsics: list, x: float, y: float, depth: float, width: int, height: int) -> list:
    """Convert point in meters into point in pixels

        Args:
            intrinsics of sensor: Tells if this is ToF sensor or RGB sensor
            x: X-pos in m
            y: Y-pos in m
            depth: distance from sensor to object at (x, y)
            width
            height

        Returns:
            tx, ty, depth
        """
    fx = intrinsics[0] * float(width)
    fy = intrinsics[1] * float(height)
    cx = intrinsics[2] * float(width)
    cy = intrinsics[3] * float(height)
    tx = x * fx / depth + cx
    ty = y * fy / depth + cy
    return [tx, ty, depth]


def parse_calibration(filepath: str) -> List[List[float]]:
    """Parse calibration file
    filepath: The content of a calibration file looks like this:
        Color camera intrinsic:
        0.6786797 0.90489584 0.49585155 0.5035042
        Depth camera intrinsic:
        0.6786797 0.90489584 0.49585155 0.5035042
    """
    with open(filepath, 'r') as f:
        calibration = []
        for _ in range(2):
            f.readline().strip()
            line_with_numbers = f.readline()
            intrinsic = parse_numbers(line_with_numbers)
            calibration.append(intrinsic)
    return calibration


def is_google_tango_resolution(width, height):
    """Check for special case for Google Tango devices with different rotation"""
    return width == 180 and height == 135


def normalize(vectors: np.ndarray) -> np.ndarray:
    """Ensure the normal has a length of one

    This way of normalizing is commonly used for normals.
    It achieves that normals are of size 1.

    Args:
        vectors (np.array): Multiple vectors (e.g. could be normals)

    Returns:
        This achieves: abs(x) + abs(y) + abs(z) = 1
    """
    length = abs(vectors[0]) + abs(vectors[1]) + abs(vectors[2])
    length[length == 0] = 1
    return vectors / length



import logging
import logging.config
import re

import numpy as np

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def get_timestamps_from_rgb(rgb_paths):
    path = [x for x in rgb_paths]
    timestamps = []

    for p in path:
        filename = p.split('/')[-1]
        value = filename.split('_')[-1]
        timestamp_value = value.replace('.jpg', '')
        timestamps.append(float(timestamp_value))

    if len(timestamps) == 0:
        error = np.array([])
        return [error, path]

    timestamps = np.asarray(timestamps)

    return [timestamps, path]


def get_timestamp_from_pcd(pcd_path):
    filename = str(pcd_path)
    infile = open(filename, 'r')
    first_line = infile.readline()

    # get the time from the header of the pcd file
    timestamp = re.findall(r"\d+\.\d+", first_line)

    # check if a timestamp is parsed from the header of the pcd file
    try:
        return_timestamp = float(timestamp[0])
    except IndexError:
        return_timestamp = []

    return return_timestamp


def get_timestamps_from_pcd(pcd_paths):
    timestamps = np.array([])
    path = [x for x in pcd_paths]

    #iterate over all paths pointing to pcds
    for p in path:
        try:
            stamp = get_timestamp_from_pcd(p)
            timestamps = np.append(timestamps, stamp)
        except IndexError:
            error = np.array([])
            logger.error("Error with timestamp in pcd")
            return [error, p]

    if len(timestamps) == 0:
        error = np.array([])
        return [error, path]
    return [timestamps, path]



import os
import re
import logging
import logging.config
from functools import partial
from itertools import groupby, islice
from typing import Iterator, List

from bunch import Bunch
import glob2 as glob

from cgmml.common.model_utils.model_utils_constants import SAMPLING_STRATEGY_SYSTEMATIC, SAMPLING_STRATEGY_WINDOW

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

REGEX_PICKLE = re.compile(
    r"pc_(?P<qrcode>[a-zA-Z0-9]+-[a-zA-Z0-9]+)_(?P<unixepoch>\d+)_(?P<code>\d{3})_(?P<idx>\d{3}).p$"
)


def create_multiartifact_paths_for_qrcodes(qrcode_paths: List[str], data_config: Bunch) -> List[List[str]]:
    samples = []
    for qrcode_path in sorted(qrcode_paths):
        for code in data_config.CODES:
            p = os.path.join(qrcode_path, code)
            new_samples = _create_multiartifact_paths(p, data_config.N_ARTIFACTS, data_config)
            samples.extend(new_samples)
    return samples


def _create_multiartifact_paths(qrcode_path: str, n_artifacts: int, data_config: Bunch) -> List[List[str]]:
    """Look at files for 1 qrcode and divide into samples.

    Args:
        qrcode_path: File path of 1 qrcode, e.g. "dataset/scans/1583462470-16tvfmb1d0/100"
        n_artifacts: Desired number of artifacts in one sample

    Returns:
        List of samples, where each sample consists of muliple file paths
    """
    path_with_wildcard = os.path.join(qrcode_path, "*.p")
    list_of_pickle_file_paths = sorted(glob.glob(path_with_wildcard))

    # Split if there are multiple scans on different days
    scans = [list(v) for _unixepoch, v in groupby(list_of_pickle_file_paths, _get_epoch) if _unixepoch]

    # Filter to keep scans with enough artifacts
    scans = list(filter(lambda x: len(x) >= n_artifacts, scans))

    # Sample artifacts
    if data_config.SAMPLING_STRATEGY == SAMPLING_STRATEGY_SYSTEMATIC:
        samples = list(map(partial(sample_systematic_from_artifacts, n_artifacts=n_artifacts), scans))

    if data_config.SAMPLING_STRATEGY == SAMPLING_STRATEGY_WINDOW:
        samples = []
        for scan in scans:
            some_samples = list(sample_windows_from_artifacts(scan, n_artifacts=n_artifacts))
            assert len(scan) - n_artifacts + 1 == len(some_samples)
            samples.extend(some_samples)

    return samples


def sample_windows_from_artifacts(artifacts: list, n_artifacts: int) -> Iterator[list]:
    """Sample multiple windows (of length n_artifacts) from list of artifacts

    Args:
        artifacts: e.g. ['001.p', '002.p', '003.p', '004.p', '005.p', '006.p']
        n_artifacts: Desired number of artifacts in one sample

    Returns:
        samples: e.g. [
            ['001.p', '002.p', '003.p', '004.p', '005.p'],
            ['002.p', '003.p', '004.p', '005.p', '006.p'],
        ]
    """
    it = iter(artifacts)
    result = list(islice(it, n_artifacts))
    if len(result) == n_artifacts:
        yield result
    for elem in it:
        result = result[1:] + [elem]
        yield result


def sample_systematic_from_artifacts(artifacts: list, n_artifacts: int) -> list:
    n_artifacts_total = len(artifacts)
    n_skip = n_artifacts_total // n_artifacts  # 20 / 5 = 4
    indexes_to_select = list(range(n_skip // 2, n_artifacts_total, n_skip))[:n_artifacts]
    selected_artifacts = [artifacts[i] for i in indexes_to_select]
    assert len(selected_artifacts) == n_artifacts, str(artifacts)
    return selected_artifacts


def _get_epoch(fname: str) -> str:
    match_result = REGEX_PICKLE.search(fname)
    if match_result:
        return match_result.group("unixepoch")
    else:
        logger.info("%s doesn't match REGEX_PICKLE", fname)



import json

import requests
from azureml.core import Webservice, Workspace

from config_deepensemble_1 import CONFIG

from cgmml.common.data_utilities import mlpipeline_utils

if __name__ == "__main__":
    if CONFIG.LOCALTEST:
        uri = 'http://localhost:6789/'
    else:
        workspace = Workspace.from_config()
        service = Webservice(workspace=workspace, name=CONFIG.ENDPOINT_NAME)
        uri = service.scoring_uri

    requests.get(uri)
    depthmap = mlpipeline_utils.get_depthmaps(CONFIG.TEST_FILES).tolist()  # Make JSON serializable

    headers = {"Content-Type": "application/json"}
    data = {
        "data": depthmap,
    }

    data = json.dumps(data)

    response = requests.post(uri, data=data, headers=headers)
    print(response.json())



'''
This file contains a data-generator that works on the data as it is provided by the ETL.
'''

from __future__ import absolute_import
import os
import logging
import logging.config
import numpy as np
import glob2 as glob
import random
import uuid
import progressbar
import multiprocessing as mp
from pyntcloud import PyntCloud
import pickle
from . import utils

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


class ETLDataGenerator(object):
    """
    This class generates data for training.
    """

    def __init__(
        self,
        dataset_path,
        input_type,
        #output_targets,
        sequence_length=0,
        image_target_shape=(160, 90),
        voxelgrid_target_shape=(32, 32, 32),
        voxel_size_meters=0.01,
        voxelgrid_random_rotation=False,
        pointcloud_target_size=32000,
        pointcloud_random_rotation=False
    ):
        """
        Initializes a DataGenerator.

        Args:
            dataset_path (string): Where the raw data is.
            input_type (string): Specifies how the input-data for the Neural Network looks like.
                                 Either 'image', 'pointcloud', 'voxgrid'.
            output_targets (list of strings): A list of targets for the Neural Network.
                                 For example *['height', 'weight']*.
            sequence_length (int): Specifies the lenght of the sequences. 0 would yield no sequence at all.
            image_target_shape (2D tuple of ints): Target shape of the images.
            voxelgrid_target_shape (3D tuple of ints): Target shape of the voxelgrids.
            voxel_size_meters (float): Size of the voxels. That is, edge length.
            voxelgrid_random_rotation (bool): If True voxelgrids will be rotated randomly.
            pointcloud_target_size (int): Target size of the pointclouds.
            pointcloud_random_rotation (bool): If True pointclouds will be rotated randomly.

        """

        # Preconditions.
        assert os.path.exists(dataset_path), "dataset_path must exist: " + str(dataset_path)
        assert isinstance(input_type, str), "input_type must be string: " + str(input_type)
        #assert isinstance(output_targets, list), "output_targets must be list: " + str(output_targets)
        if input_type == "image":
            assert len(image_target_shape) == 2, "image_target_shape must be 2-dimensional: " + \
                str(image_target_shape)
        if input_type == "voxelgrid":
            assert len(voxelgrid_target_shape) == 3, "voxelgrid_target_shape must be 3-dimensional: " + \
                str(voxelgrid_target_shape)

        # Assign the instance-variables.
        self.dataset_path = dataset_path
        self.input_type = input_type
        #self.output_targets = output_targets
        self.sequence_length = sequence_length
        self.image_target_shape = image_target_shape
        self.voxelgrid_target_shape = voxelgrid_target_shape
        self.voxel_size_meters = voxel_size_meters
        self.voxelgrid_random_rotation = voxelgrid_random_rotation
        self.pointcloud_target_size = pointcloud_target_size
        self.pointcloud_random_rotation = pointcloud_random_rotation

        # Find all QR-codes.
        self._find_qrcodes()
        assert self.qrcodes is not None, "No QR-codes found!"

        # Prepare the data.
        self._prepare_qrcodes_dictionary()

        # Create some caches.
        self.image_cache = {}
        self.voxelgrid_cache = {}
        self.pointcloud_cache = {}

        # Check if paths are fine.
        if self.input_type == "image":
            assert self.all_jpg_paths != []
        elif self.input_type == "voxelgrid" or self.input_type == "pointcloud":
            assert self.all_pcd_paths != []
        else:
            raise Exception("Unexpected: " + self.input_type)

    def _find_qrcodes(self):
        """
        Finds all QR-codes.

        Each individual is represented via a unique QR-code. This method extracts the set of QR-codes.
        """

        # Retrieve the QR-codes from the folders.
        paths = glob.glob(os.path.join(self.dataset_path, "*"))
        self.qrcodes = sorted([path.split("/")[-1] for path in paths])

    def _prepare_qrcodes_dictionary(self):

        self.all_pcd_paths = []
        self.all_jpg_paths = []

        self.qrcodes_dictionary = {}
        for qrcode in self.qrcodes:
            self.qrcodes_dictionary[qrcode] = []
            #logger.info("QR-code:", qrcode)
            measurement_paths = glob.glob(os.path.join(self.dataset_path, qrcode, "*"))
            for measurement_path in measurement_paths:
                # Getting PCDs.
                pcd_paths = glob.glob(os.path.join(measurement_path, "pcd", "*.pcd"))

                # Getting JPGs.
                jpg_paths = glob.glob(os.path.join(measurement_path, "jpg", "*.jpg"))

                # Loading the targets.
                target_path = os.path.join(measurement_path, "target.txt")
                target_file = open(target_path, "r")
                targets = np.array([float(value) for value in target_file.read().split(",")])
                target_file.close()

                # Done.
                self.qrcodes_dictionary[qrcode].append((pcd_paths, jpg_paths, targets))
                self.all_pcd_paths.extend(pcd_paths)
                self.all_jpg_paths.extend(jpg_paths)

    def analyze_files(self):

        for qrcode in self.qrcodes:
            logger.info("QR-code: %s", qrcode)
            for pcd_paths, jpg_paths, targets in self.qrcodes_dictionary[qrcode]:
                logger.info(" %d PCD-files for targets %s", len(pcd_paths), targets)
                logger.info(" %d JPG-files for targets %s", len(jpg_paths), targets)

        logger.info("Total PCD-files: %d", len(self.all_pcd_paths))
        logger.info("Total JPG-files: %d", len(self.all_jpg_paths))

        #measurement_paths = glob.glob(os.path.join(self.dataset_path, qrcode, "*"))
        #for measurement_path in measurement_paths:
        #    logger.info("  %s", measurement_path.split("/")[-1])
        #    pcd_paths = glob.glob(os.path.join(measurement_path, "*.pcd"))
        #    logger.info("    %d PCDs", len(pcd_paths))
        #logger.info("Number of JPGs: %d", len(self.jpg_paths))
        #logger.info("Number of PCDs: %d", len(self.pcd_paths))
        #logger.info("Number of JSONs (personal): %d", len(self.json_paths_personal))
        #logger.info("Number of JSONs (measures): %d", len(self.json_paths_measures))

    def generate(self, size, qrcodes_to_use=None, verbose=False, yield_file_paths=False, multiprocessing_jobs=1):

        if qrcodes_to_use is None:
            qrcodes_to_use = self.qrcodes

        # Main loop.
        while True:

            # Use only a single process.
            if multiprocessing_jobs == 1:
                yield generate_data(self, size, qrcodes_to_use, verbose, yield_file_paths, None)

            # Use multiple processes.
            elif multiprocessing_jobs > 1:

                # Create chunks of almost equal size.
                subset_sizes = [0] * multiprocessing_jobs
                subset_sizes[0:multiprocessing_jobs - 1] = [size
                                                            // multiprocessing_jobs] * (multiprocessing_jobs - 1)
                subset_sizes[multiprocessing_jobs - 1] = size - \
                    sum(subset_sizes[0:multiprocessing_jobs - 1])
                subset_sizes = [s for s in subset_sizes if s > 0]
                assert sum(subset_sizes) == size

                # Create an output_queue.
                output_queue = mp.Queue()

                # Create the processes.
                processes = []
                for subset_size in subset_sizes:
                    process_target = generate_data
                    process_args = (self, subset_size, qrcodes_to_use,
                                    verbose, yield_file_paths, output_queue)
                    process = mp.Process(target=process_target, args=process_args)
                    processes.append(process)

                # Start the processes.
                for p in processes:
                    p.start()

                # Exit the completed processes.
                for p in processes:
                    p.join()

                # Get process results from the output queue
                output_paths = [output_queue.get() for p in processes]

                # Merge data.
                x_inputs_arrays = []
                y_outputs_arrays = []
                file_paths_arrays = []
                for output_path in output_paths:
                    # Read data from file and delete it.
                    result_values = pickle.load(open(output_path, "rb"))
                    assert len(result_values[0]) > 0
                    assert len(result_values[1]) > 0
                    os.remove(output_path)

                    # Gather the data into arrays.
                    if x_inputs_arrays != []:
                        assert result_values[0].shape[1:] == x_inputs_arrays[-1].shape[1:], str(
                            result_values[0].shape) + " vs " + str(x_inputs_arrays[-1].shape)
                    if y_outputs_arrays != []:
                        assert result_values[1].shape[1:] == y_outputs_arrays[-1].shape[1:], str(
                            result_values[1].shape) + " vs " + str(y_outputs_arrays[-1].shape)
                    x_inputs_arrays.append(result_values[0])
                    y_outputs_arrays.append(result_values[1])
                    if yield_file_paths is True:
                        file_paths_arrays.append(result_values[2])
                    else:
                        file_paths_arrays.append([])

                x_inputs = np.concatenate(x_inputs_arrays)
                y_outputs = np.concatenate(y_outputs_arrays)
                file_paths = np.concatenate(file_paths_arrays)
                assert len(x_inputs) == size
                assert len(y_outputs) == size

                # Done.
                if yield_file_paths is False:
                    yield x_inputs, y_outputs
                else:
                    yield x_inputs, y_outputs, file_paths

            else:
                raise Exception("Unexpected value for 'multiprocessing_jobs' "
                                + str(multiprocessing_jobs))

    def _load_pointcloud(self, pcd_path, preprocess=True, augmentation=True):

        pointcloud = self.pointcloud_cache.get(pcd_path, [])
        if pointcloud is None:
            pointcloud = PyntCloud.from_file(pcd_path).points.values

            if self.pointcloud_target_size is not None and preprocess is True:
                pointcloud = np.array(pointcloud)[:, 0:3]  # Drop confidence.
                pointcloud = pointcloud[:self.pointcloud_target_size]
                if len(pointcloud) < self.pointcloud_target_size:
                    zeros = np.zeros((self.pointcloud_target_size - len(pointcloud), 4))
                    pointcloud = np.concatenate([pointcloud, zeros])

            if self.pointcloud_random_rotation is True and augmentation is True:
                numpy_points = pointcloud[:, 0:3]
                numpy_points = self._rotate_point_cloud(numpy_points)
                pointcloud[:, 0:3] = numpy_points

            #self.pointcloud_cache[pcd_path] = pointcloud
        return pointcloud

    def _load_voxelgrid(self, pcd_path, preprocess=True, augmentation=True):
        voxelgrid = self.voxelgrid_cache.get(pcd_path, [])
        if voxelgrid == []:

            # Load the pointcloud.
            point_cloud = PyntCloud.from_file(pcd_path)
            if self.voxelgrid_random_rotation is True and augmentation is True:
                points = point_cloud.points
                numpy_points = points.values[:, 0:3]
                numpy_points = self._rotate_point_cloud(numpy_points)
                points.iloc[:, 0:3] = numpy_points
                point_cloud.points = points

            # Create voxelgrid from pointcloud.
            voxelgrid_id = point_cloud.add_structure(
                "voxelgrid",
                size_x=self.voxel_size_meters,
                size_y=self.voxel_size_meters,
                size_z=self.voxel_size_meters)
            voxelgrid = point_cloud.structures[voxelgrid_id].get_feature_vector(mode="density")

            # Do the preprocessing.
            if preprocess is True:
                voxelgrid = utils.ensure_voxelgrid_shape(voxelgrid, self.voxelgrid_target_shape)
                assert voxelgrid.shape == self.voxelgrid_target_shape

            #self.voxelgrid_cache[pcd_path] = voxelgrid # TODO cache is turned off because of you know why...

        return voxelgrid

    def _rotate_point_cloud(self, point_cloud):

        rotation_angle = np.random.uniform() * 2 * np.pi
        cosval = np.cos(rotation_angle)
        sinval = np.sin(rotation_angle)
        rotation_matrix = np.array([[cosval, sinval, 0],
                                    [-sinval, cosval, 0],
                                    [0, 0, 1]])

        rotated_data = np.zeros(point_cloud.shape, dtype=np.float32)
        for k in range(point_cloud.shape[0]):

            shape_pc = point_cloud[k, ...]
            rotated_data[k, ...] = np.dot(shape_pc.reshape((-1, 3)), rotation_matrix)

        return rotated_data


def create_datagenerator_from_parameters(dataset_path, dataset_parameters):
    logger.info("Creating data-generator...")
    datagenerator = ETLDataGenerator(
        dataset_path=dataset_path,
        input_type=dataset_parameters["input_type"],
        #output_targets=dataset_parameters["output_targets"],
        sequence_length=dataset_parameters.get("sequence_length", 0),
        voxelgrid_target_shape=dataset_parameters.get("voxelgrid_target_shape", None),
        voxel_size_meters=dataset_parameters.get("voxel_size_meters", None),
        voxelgrid_random_rotation=dataset_parameters.get("voxelgrid_random_rotation", None),
        pointcloud_target_size=dataset_parameters.get("pointcloud_target_size", None),
        pointcloud_random_rotation=dataset_parameters.get("pointcloud_random_rotation", None)
    )
    #datagenerator.print_statistics()
    return datagenerator


def get_dataset_path(root_path="../data/etl"):
    if os.path.exists("etldatasetpath.txt"):
        with open("etldatasetpath.txt", "r") as file:
            dataset_path = file.read().replace("\n", "")
    else:
        # Finding the latest.
        dataset_paths = glob.glob(os.path.join(root_path, "*"))
        dataset_paths = [
            dataset_path for dataset_path in dataset_paths if os.path.isdir(dataset_path)]
        dataset_path = list(reversed(sorted(dataset_paths)))[0]

    return dataset_path


def generate_data(class_self, size, qrcodes_to_use, verbose, yield_file_paths, output_queue):
    if verbose is True:
        logger.info("Generating QR-codes to be used: %s", qrcodes_to_use)

    assert size != 0

    x_inputs = []
    y_outputs = []
    file_paths = []

    if verbose is True:
        bar = progressbar.ProgressBar(max_value=size)
    while len(x_inputs) < size:

        # Get a random QR-code.
        qrcode = random.choice(qrcodes_to_use)

        # Get targets and paths randomly.
        if qrcode not in class_self.qrcodes_dictionary.keys():
            continue
        pcd_paths, jpg_paths, targets = random.choice(class_self.qrcodes_dictionary[qrcode])

        # No pointclouds.
        if len(pcd_paths) == 0:
            continue

        # Get a sample.
        x_input = None
        y_output = None
        file_path = None

        # Get the input. Not dealing with sequences.
        if class_self.sequence_length == 0:
            x_input, file_path = get_input(class_self, jpg_paths, pcd_paths)

        # Get the input. Dealing with sequences here.
        else:
            count = 0
            x_input, file_path = [], []
            while count != class_self.sequence_length:
                x, f = get_input(class_self, jpg_paths, pcd_paths)
                if x is not None and f is not None:
                    x_input.append(x)
                    file_path.append(f)
                    count += 1
            x_input = np.array(x_input)
            file_path = np.array(file_path)

        # Set the output.
        y_output = targets

        # Got a proper sample.
        if x_input is not None and y_output is not None and file_path is not None:
            x_inputs.append(x_input)
            y_outputs.append(y_output)
            file_paths.append(file_path)

        assert len(x_inputs) == len(y_outputs)
        assert len(y_outputs) == len(file_paths)

        if verbose is True:
            bar.update(len(x_inputs))

    if verbose is True:
        bar.finish()

    assert len(x_inputs) == size
    assert len(y_outputs) == size

    # Turn everything into ndarrays.
    x_inputs = np.array(x_inputs)
    y_outputs = np.array(y_outputs)

    # Prepare result values.
    assert len(x_inputs) == size
    assert len(y_outputs) == size
    if yield_file_paths is False:
        return_values = (x_inputs, y_outputs)
    else:
        return_values = (x_inputs, y_outputs, file_paths)

    # This is used in multiprocessing. Creates a pickle file and puts the data there.
    if output_queue is not None:
        output_path = uuid.uuid4().hex + ".p"
        pickle.dump(return_values, open(output_path, "wb"))
        output_queue.put(output_path)
    else:
        return return_values


def get_input(class_self, jpg_paths, pcd_paths):
    # Get a random image.
    if class_self.input_type == "image":
        if len(jpg_paths) == 0:
            logger.info("777")
            return None, None
        jpg_path = random.choice(jpg_paths)
        image = class_self._load_image(jpg_path)
        file_path = jpg_path
        x_input = image

    # Get a random voxelgrid.
    elif class_self.input_type == "voxelgrid":
        if len(pcd_paths) == 0:
            return None, None
        pcd_path = random.choice(pcd_paths)
        try:
            voxelgrid = class_self._load_voxelgrid(pcd_path)
            file_path = pcd_path
            x_input = voxelgrid
        except Exception as e:
            print(e)
            return None, None

    # Get a random pointcloud.
    elif class_self.input_type == "pointcloud":
        if len(pcd_paths) == 0:
            return None, None
        pcd_path = random.choice(pcd_paths)
        try:
            pointcloud = class_self._load_pointcloud(pcd_path)
            file_path = pcd_path
            x_input = pointcloud
        except Exception as e:
            print(e)
            return None, None

    # Should not happen.
    else:
        raise Exception("Unknown input_type: " + class_self.input_type)

    return x_input, file_path



import zipfile

import numpy as np

from typing import List
from cgmml.common.depthmap_toolkit.depthmap import convert_3d_to_2d
from cgmml.common.depthmap_toolkit.depthmap_utils import parse_numbers

ENCODING = 'charmap'


def parse_pcd(filepath: str) -> List[List[float]]:
    with open(filepath, 'r') as f:
        data = []
        while True:
            line = str(f.readline())
            if line.startswith('DATA'):
                break

        while True:
            line = str(f.readline())
            if not line:
                break
            else:
                values = parse_numbers(line)
                data.append(values)
    return data


def process(calibration, pcd_fpath: str, width: int, height: int) -> np.ndarray:
    # Convert to depthmap
    points = parse_pcd(pcd_fpath)
    output = np.zeros((width, height, 3))
    for p in points:
        v = convert_3d_to_2d(calibration[1], p[0], p[1], p[2], width, height)
        x = int(width - v[0] - 1)
        y = int(height - v[1] - 1)
        if x >= 0 and y >= 0 and x < width and y < height:
            output[x, y, 0] = p[3]
            output[x, y, 2] = p[2]
    return output


def write_depthmap(output_depth_fpath: str, depthmap: np.ndarray, width: int, height: int):
    # Write depthmap
    with open('data', 'wb') as f:
        header_str = str(width) + 'x' + str(height) + '_0.001_255\n'
        f.write(header_str.encode(ENCODING))
        for y in range(height):
            for x in range(width):
                depth = int(depthmap[x, y, 2] * 1000)
                confidence = int(depthmap[x, y, 0] * 255)
                depth_byte = chr(int(depth / 256)).encode(ENCODING)
                depth_byte2 = chr(depth % 256).encode(ENCODING)
                confidence_byte = chr(confidence).encode(ENCODING)
                f.write(depth_byte)
                f.write(depth_byte2)
                f.write(confidence_byte)
    # Zip data
    with zipfile.ZipFile(output_depth_fpath, "w", zipfile.ZIP_DEFLATED) as f:
        f.write('data', 'data')
        f.close()



import logging
import logging.config
import os
import pickle
from multiprocessing import Pool
from pathlib import Path
import time
from typing import Callable, List, Tuple

import glob2 as glob
import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras.models import load_model
from azureml.core import Experiment, Run
from bunch import Bunch
from scipy.stats.stats import pearsonr
from sklearn.metrics import ConfusionMatrixDisplay, confusion_matrix
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt  # noqa: E402
from cgmzscore import Calculator  # noqa: E402

from .constants_eval import (  # noqa: E402
    CODE_TO_SCANTYPE, COLUMN_NAME_AGE, COLUMN_NAME_GOODBAD, COLUMN_NAME_SEX, DAYS_IN_YEAR,
    GOODBAD_DICT, SEX_DICT)
from .eval_utils import (  # noqa: E402
    avgerror, preprocess_depthmap, preprocess_targets)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

MIN_HEIGHT = 45
MAX_HEIGHT = 120
MAX_AGE = 1856.0

STUNTING_DIAGNOSIS = ["Not Stunted", "Moderately Stunted", "Severly Stunted"]
WASTING_DIAGNOSIS = ["Not Under-weight", "Moderately Under-weight", "Severly Under-weight"]


def process_image(data):
    img = tf.convert_to_tensor(data)
    img = tf.cast(img, tf.float32) * (1. / 256)
    img = tf.image.rot90(img, k=3)
    img = tf.image.resize(img, [240, 180])
    img = tf.expand_dims(img, axis=0)
    return img


def get_depthmap_files(paths: List[str]) -> List[str]:
    """Prepare the list of all the depthmap pickle files in dataset"""
    pickle_paths = []
    for path in paths:
        pickle_paths.extend(glob.glob(os.path.join(path, "**", "*.p")))
    return pickle_paths


def get_column_list(depthmap_path_list: List[str], prediction: np.ndarray, data_config: Bunch):
    """Prepare the list of all artifact with its corresponding scantype, qrcode, target and prediction"""
    qrcode_list, scan_type_list, artifact_list, prediction_list, target_list = [], [], [], [], []

    for idx, path in enumerate(depthmap_path_list):
        loaded_tuple = pickle.load(open(path, "rb"))  # tuple can have 2 or 3 elements
        targets = loaded_tuple[1]
        targets = preprocess_targets(targets, data_config.TARGET_INDEXES)
        target = np.squeeze(targets)

        sub_folder_list = path.split('/')
        qrcode_list.append(sub_folder_list[-3])
        scan_type_list.append(sub_folder_list[-2])
        artifact_list.append(sub_folder_list[-1])
        prediction_list.append(prediction[idx])
        target_list.append(target)

    return qrcode_list, scan_type_list, artifact_list, prediction_list, target_list


def calculate_and_save_results(df_grouped: pd.DataFrame,
                               complete_name: str,
                               csv_out_fpath: str,
                               data_config: Bunch,
                               result_config: Bunch,
                               fct: Callable):
    """Calculate accuracies across the scantypes and save the final results table to the CSV file

    Args:
        df_grouped: dataframe grouped by 'qrcode' and 'scantype
        complete_name: e.g. 'q3-depthmap-plaincnn-height-100-95k-run_17'
        csv_out_fpath: CSV output path
        data_config: bunch containing data config
        result_config: bunch containing result config
        fct: Function to execute on inputs
    """
    dfs = []
    for code in data_config.CODES:
        df = fct(code, df_grouped, result_config)
        full_model_name = complete_name + CODE_TO_SCANTYPE[code]
        df.rename(index={0: full_model_name}, inplace=True)
        dfs.append(df)

    result = pd.concat(dfs)
    result.index.name = 'Model_Scantype'
    result = result.round(2)
    # Save the model results in csv file
    Path(csv_out_fpath).parent.mkdir(parents=True, exist_ok=True)
    result.to_csv(csv_out_fpath, index=True)


def calculate_performance_sex(code: str, df_mae: pd.DataFrame, result_config: Bunch) -> pd.DataFrame:
    df_mae_filtered = df_mae.iloc[df_mae.index.get_level_values('scantype') == code]
    accuracy_list = calculate_accuracies(SEX_DICT.values(),
                                         df_mae_filtered,
                                         COLUMN_NAME_SEX,
                                         result_config.ACCURACY_MAIN_THRESH)
    df_out = pd.DataFrame(accuracy_list)
    df_out = df_out.T
    df_out.columns = SEX_DICT.keys()
    return df_out


def calculate_performance_goodbad(code: str, df_mae: pd.DataFrame, result_config: Bunch) -> pd.DataFrame:
    df_mae_filtered = df_mae.iloc[df_mae.index.get_level_values('scantype') == code]
    accuracy_list = calculate_accuracies(GOODBAD_DICT.values(),
                                         df_mae_filtered,
                                         COLUMN_NAME_GOODBAD,
                                         result_config.ACCURACY_MAIN_THRESH)
    df_out = pd.DataFrame(accuracy_list)
    df_out = df_out.T
    df_out.columns = GOODBAD_DICT.keys()
    return df_out


def calculate_performance_age(code: str, df_mae: pd.DataFrame, result_config: Bunch) -> pd.DataFrame:
    df_mae_filtered = df_mae.iloc[df_mae.index.get_level_values('scantype') == code]

    age_thresholds = result_config.AGE_BUCKETS
    age_buckets = list(zip(age_thresholds[:-1], age_thresholds[1:]))

    accuracy_list = calculate_accuracies_on_age_buckets(age_buckets,
                                                        df_mae_filtered,
                                                        COLUMN_NAME_AGE,
                                                        result_config.ACCURACY_MAIN_THRESH)
    df_out = pd.DataFrame(accuracy_list)
    df_out = df_out.T
    df_out.columns = [f"{age_min} to {age_max}" for age_min, age_max in age_buckets]
    return df_out


def calculate_accuracies(values_to_select: List[float],
                         df: pd.DataFrame,
                         column_name: str,
                         accuracy_thresh: float) -> List[float]:
    """Take a dataframe with evaluation results and calculate cases above a threshold

    Args:
        values_to_select: Values that a certain column can have
        df: Needs to at least have to columns: 'error' and column_name
        column_name: Name of the column to select on
        accuracy_thresh: Error threshold

    Returns:
        A list of accuracies which has as many items as values_to_select
    """
    accuracy_list = []
    for idx in values_to_select:
        selection = (df[column_name] == idx)
        df_selected = df[selection]

        selection = (df_selected['error'] <= accuracy_thresh) & (df_selected['error'] >= -accuracy_thresh)
        accuracy = calc_accuracy_in_percent(num_all=len(df_selected), num_good=len(df_selected[selection]))
        accuracy_list.append(accuracy)
    return accuracy_list


def calculate_accuracies_on_age_buckets(age_buckets: List[Tuple[int]],
                                        df: pd.DataFrame,
                                        column_name: str,
                                        accuracy_thresh: float) -> List[float]:
    """Take a dataframe with evaluation results and calculate cases above a threshold

    Args:
        age_buckets: List of tuples where each tuple specifies a range: [age_min, age_max)
        df: Needs to at least have to columns: 'error' and column_name
        column_name: Name of the column to select on
        accuracy_thresh: Error threshold

    Returns:
        A list of accuracies which has as many items as values_to_select
    """
    accuracy_list = []
    for age_min_years, age_max_years in age_buckets:
        age_min = age_min_years * DAYS_IN_YEAR
        age_max = age_max_years * DAYS_IN_YEAR

        selection = (df[column_name] >= age_min) & (df[column_name] < age_max)
        df_selected = df[selection]

        selection = (df_selected['error'] <= accuracy_thresh) & (df_selected['error'] >= -accuracy_thresh)
        accuracy = calc_accuracy_in_percent(num_all=len(df_selected), num_good=len(df_selected[selection]))
        accuracy_list.append(accuracy)
    return accuracy_list


def calc_accuracy_in_percent(num_all: int, num_good: int) -> float:
    assert num_all >= num_good, f"num_all smaller than num_good: {num_good} < {num_all}"
    if num_all > 0:
        return num_good / num_all * 100
    return 0.


def draw_uncertainty_goodbad_plot(df_: pd.DataFrame, png_out_fpath: str):
    """Take all good samples and plot error distributions. Do the same for bad samples.

    Args:
        df: Dataframe with columns: goodbad and uncertainties
        png_out_fpath (str): File path where plot image will be saved
    """
    df = df_[df_.uncertainties.notna()]
    df_good = df[df[COLUMN_NAME_GOODBAD] == 1.0]
    df_bad = df[df[COLUMN_NAME_GOODBAD] == 0.0]

    good = list(df_good.uncertainties)
    bad = list(df_bad.uncertainties)

    bins = np.linspace(0, 10, 30)

    plt.hist(good, bins, alpha=0.5, label='good')
    plt.hist(bad, bins, alpha=0.5, label='bad')
    plt.title(f"Uncertainty plot: n_good={len(good)}, n_bad={len(bad)}")
    plt.xlabel("uncertainty in cm")
    plt.ylabel("occurrence count")
    plt.legend(loc='upper right')

    mean_good = float(df_good.uncertainties.mean())
    mean_bad = float(df_bad.uncertainties.mean())
    plt.axvline(mean_good, color='g', linestyle='dashed', linewidth=2)
    plt.axvline(mean_bad, color='r', linestyle='dashed', linewidth=2)

    plt.savefig(png_out_fpath)
    plt.close()


def draw_age_scatterplot(df_: pd.DataFrame, png_out_fpath: str):
    """Draw error over age scatterplot

    Args:
        df_: Dataframe with columns: qrcode, scantype, COLUMN_NAME_AGE, GT, predicted
        png_out_fpath: File path where plot image will be saved
    """
    df = df_[df_.scantype == '100'].groupby('qrcode').mean()
    df['error'] = df.apply(avgerror, axis=1).abs()
    plt.scatter(df[COLUMN_NAME_AGE], df['error'], s=2)
    plt.grid()
    plt.title("Per-scan Error over Age")
    plt.xlabel("age")
    plt.ylabel("error")
    axes = plt.gca()
    axes.set_xlim([0, 2500])
    axes.set_ylim([0, 5])
    Path(png_out_fpath).parent.mkdir(parents=True, exist_ok=True)
    plt.savefig(png_out_fpath)
    plt.close()


def draw_uncertainty_scatterplot(df: pd.DataFrame, png_out_fpath: str):
    """Draw error over age scatterplot

    Args:
        df: Dataframe with columns: qrcode, scantype, COLUMN_NAME_AGE, GT, predicted
        png_out_fpath: File path where plot image will be saved
    """
    df['error'] = df.apply(avgerror, axis=1).abs()
    plt.scatter(df['error'], df['uncertainties'], s=2)
    plt.grid()

    correlation, _ = pearsonr(df['error'], df['uncertainties'])
    logger.info("correlation: %d", correlation)

    plt.title(f"Per-scan sample artifact: Error over uncertainty (correlation={correlation:.3})")
    plt.xlabel("error")
    plt.ylabel("uncertainty (stdev of MC Dropout)")
    axes = plt.gca()
    axes.set_xlim([0, 5])
    axes.set_ylim([0, 10])
    Path(png_out_fpath).parent.mkdir(parents=True, exist_ok=True)
    plt.savefig(png_out_fpath)
    plt.close()


def draw_stunting_diagnosis(df: pd.DataFrame, png_out_fpath: str):
    """Draw stunting Confusion Matrix

    Args:
        df: Dataframe with columns: qrcode, scantype, COLUMN_NAME_AGE, GT, predicted
        png_out_fpath: File path where plot image will be saved
    """
    df = parallelize_dataframe(df, calculate_zscore_lhfa)
    actual = np.where(df['Z_actual'].values < -3, 'Severly Stunted',
                      np.where(df['Z_actual'].values > -2, 'Not Stunted', 'Moderately Stunted'))
    predicted = np.where(df['Z_predicted'].values < -3, 'Severly Stunted',
                         np.where(df['Z_predicted'].values > -2, 'Not Stunted', 'Moderately Stunted'))
    data = confusion_matrix(actual, predicted)
    draw_confusion_matrix(data, png_out_fpath, STUNTING_DIAGNOSIS, 'Stunting Diagnosis')


def calculate_zscore_lhfa(df):
    """lhfa: length/height for age"""
    cal = Calculator()

    def _calc_score(age_in_days, height, sex):
        if MIN_HEIGHT < height <= MAX_HEIGHT and age_in_days <= MAX_AGE:
            return cal.zScore_lhfa(age_in_days=age_in_days, sex=sex, height=height)

    def _fct(row):
        return _calc_score(age_in_days=int(row[COLUMN_NAME_AGE]),
                           sex='M' if row[COLUMN_NAME_SEX] == SEX_DICT['male'] else 'F', height=row['GT'])

    def _fct2(row):
        return _calc_score(age_in_days=int(row[COLUMN_NAME_AGE]),
                           sex='M' if row[COLUMN_NAME_SEX] == SEX_DICT['male'] else 'F', height=row['predicted'])

    df['Z_actual'] = df.apply(_fct, axis=1)
    df['Z_predicted'] = df.apply(_fct2, axis=1)
    return df


def draw_wasting_diagnosis(df: pd.DataFrame, png_out_fpath: str):
    """Draw wasting Confusion Matrix

    Args:
        df_: Dataframe with columns: qrcode, scantype, COLUMN_NAME_AGE, GT, predicted
        png_out_fpath: File path where plot image will be saved
    """
    df = parallelize_dataframe(df, calculate_zscore_wfa)
    actual = np.where(df['Z_actual'].values < -3, 'Severly Under-weight',
                      np.where(df['Z_actual'].values > -2, 'Not Under-weight', 'Moderately Under-weight'))
    predicted = np.where(df['Z_predicted'].values < -3, 'Severly Under-weight',
                         np.where(df['Z_predicted'].values > -2, 'Not Under-weight', 'Moderately Under-weight'))
    data = confusion_matrix(actual, predicted)
    draw_confusion_matrix(data, png_out_fpath, WASTING_DIAGNOSIS, 'Wasting Diagnosis')


def calculate_zscore_wfa(df):
    """Weight for age"""
    cal = Calculator()

    def utils(age_in_days, weight, sex):
        if age_in_days <= MAX_AGE:
            return cal.zScore_wfa(age_in_days=age_in_days, sex=sex, weight=weight)

    df['Z_actual'] = df.apply(
        lambda row: utils(age_in_days=int(row[COLUMN_NAME_AGE]),
                          sex='M' if row[COLUMN_NAME_SEX] == SEX_DICT['male'] else 'F', weight=row['GT']),
        axis=1)
    df['Z_predicted'] = df.apply(
        lambda row: utils(age_in_days=int(row[COLUMN_NAME_AGE]),
                          sex='M' if row[COLUMN_NAME_SEX] == SEX_DICT['male'] else 'F', weight=row['predicted']),
        axis=1)
    return df


def draw_confusion_matrix(data, png_out_fpath, display_labels, title):
    T, FP, FN = calculate_percentage_confusion_matrix(data)
    fig = plt.figure(figsize=(15, 15))
    ax = fig.add_subplot(111)
    disp = ConfusionMatrixDisplay(confusion_matrix=data, display_labels=display_labels)
    disp.plot(cmap='Blues', values_format='d', ax=ax)
    s = f"True: {round(T, 2)} False Positive: {round(FP, 2)} False Negative: {round(FN, 2)}"
    plt.text(0.5, 0.5, s, size=10, bbox=dict(boxstyle="square", facecolor='white'))
    ax.set_title(title)
    Path(png_out_fpath).parent.mkdir(parents=True, exist_ok=True)
    plt.savefig(png_out_fpath)
    plt.close()


def parallelize_dataframe(df, calculate_confusion_matrix, n_cores=8):
    df_split = np.array_split(df, n_cores)
    pool = Pool(n_cores)
    df = pd.concat(pool.map(calculate_confusion_matrix, df_split))
    pool.close()
    pool.join()
    return df


def calculate_percentage_confusion_matrix(data):
    T1, FP1, FP2, FN1, T2, FP3, FN2, FN3, T3 = data.ravel()
    Total = sum(data.ravel())
    T = round(((T1 + T2 + T3) / Total) * 100, 2)
    FP = round(((FP1 + FP2 + FP3) / Total) * 100, 2)
    FN = round(((FN1 + FN2 + FN3) / Total) * 100, 2)
    return T, FP, FN


def get_model_path(model_config: Bunch) -> str:
    if model_config.NAME.endswith(".h5"):
        return model_config.NAME
    if model_config.NAME.endswith(".ckpt"):
        return os.path.join(model_config.INPUT_LOCATION, model_config.NAME)
    raise NameError(f"{model_config.NAME}'s path extension not supported")


def download_model(workspace, experiment_name, run_id, input_location, output_location):
    """Download the pretrained model

    Args:
         workspace: workspace to access the experiment
         experiment_name: Name of the experiment in which model is saved
         run_id: Run Id of the experiment in which model is pre-trained
         input_location: Input location in a RUN Id
         output_location: Location for saving the model
    """
    experiment = Experiment(workspace=workspace, name=experiment_name)
    # Download the model on which evaluation need to be done
    run = Run(experiment, run_id=run_id)
    if input_location.endswith(".h5"):
        run.download_file(input_location, output_location)
    elif input_location.endswith(".ckpt"):
        run.download_files(prefix=input_location, output_directory=output_location)
    else:
        raise NameError(f"{input_location}'s path extension not supported")

    logger.info("Successfully downloaded model")


def filter_dataset_according_to_standing_lying(paths_evaluation: List[str],
                                               standing_model: tf.keras.models.Model) -> List[str]:
    paths_belonging_to_predictions = []
    exc = []
    for p in paths_evaluation:
        _depthmap, _targets, image = pickle.load(open(p, "rb"))
        try:
            image = process_image(image)
            if standing_model.predict(image) > .9:
                paths_belonging_to_predictions.append(p)
        except ValueError:
            exc.append(image)
    return paths_belonging_to_predictions


def get_prediction(model_path: str, dataset_evaluation: tf.data.Dataset, data_config) -> np.ndarray:
    """Perform the prediction on the dataset with the given model.

    Args:
        model_path: Path of the trained model
        dataset_evaluation: dataset in which the evaluation need to performed
    Returns:
        predictions, array shape (N_SAMPLES, )
    """
    logger.info("loading model from %s", model_path)
    model = load_model(model_path, compile=False)

    dataset = dataset_evaluation.batch(data_config.BATCH_SIZE)

    logger.info("starting predicting")
    start = time.time()
    predictions = model.predict(dataset, batch_size=data_config.BATCH_SIZE)
    end = time.time()
    logger.info("Total time for uncertainty prediction experiment: %.2f sec", end - start)

    prediction_list = np.squeeze(predictions)
    return prediction_list


def get_predictions_from_multiple_models(model_paths: list, dataset_evaluation: tf.data.Dataset, data_config) -> list:
    prediction_array = []
    for model_index, model_path in enumerate(model_paths):
        logger.info(f"Model {model_index + 1}/{len(model_paths)}")
        prediction_array += [get_prediction(model_path, dataset_evaluation, data_config)]
        logger.info("Prediction made by model on the depthmaps...")
    prediction_array = np.array(prediction_array)
    prediction_array = np.mean(prediction_array, axis=0)
    return prediction_array


def get_prediction_multiartifact(model_path: str,
                                 dataset: tf.data.Dataset,
                                 data_config: Bunch) -> np.ndarray:
    """Make prediction on each multiartifact sample.

    Args:
        model_path: File path to the model
        samples_paths: A list of samples where each sample contains N_ARTIFACTS.
        data_config

    Returns:
        predictions array
    """
    logger.info("loading model from %s", model_path)
    model = load_model(model_path, compile=False)
    predictions = model.predict(dataset, batch_size=data_config.BATCH_SIZE)
    return predictions


def tf_load_pickle(path, max_value, data_config):
    """Utility to load the depthmap (may include RGB) pickle file"""
    def py_load_pickle(path, max_value):
        loaded_tuple = pickle.load(open(path.numpy(), "rb"))  # tuple can have 2 or 3 elements
        depthmap = loaded_tuple[0]
        targets = loaded_tuple[1]
        depthmap = preprocess_depthmap(depthmap)
        depthmap = depthmap / max_value
        depthmap = tf.image.resize(depthmap, (data_config.IMAGE_TARGET_HEIGHT, data_config.IMAGE_TARGET_WIDTH))
        targets = preprocess_targets(targets, data_config.TARGET_INDEXES)
        return depthmap, targets

    depthmap, targets = tf.py_function(py_load_pickle, [path, max_value], [tf.float32, tf.float32])
    depthmap.set_shape((data_config.IMAGE_TARGET_HEIGHT, data_config.IMAGE_TARGET_WIDTH, 1))
    targets.set_shape((len(data_config.TARGET_INDEXES,)))
    return path, depthmap, targets


def prepare_sample_dataset(df_sample: pd.DataFrame, dataset_path: str, data_config: Bunch) -> tf.data.Dataset:
    df_sample['artifact_path'] = df_sample.apply(
        lambda x: f"{dataset_path}/scans/{x['qrcode']}/{x['scantype']}/{x['artifact']}", axis=1)
    paths_evaluation = list(df_sample['artifact_path'])
    dataset_sample = tf.data.Dataset.from_tensor_slices(paths_evaluation)
    dataset_sample = dataset_sample.map(
        lambda path: tf_load_pickle(path, data_config.NORMALIZATION_VALUE, data_config)
    )
    dataset_sample = dataset_sample.map(lambda _path, depthmap, targets: (depthmap, targets))
    dataset_sample = dataset_sample.cache()
    dataset_sample = dataset_sample.prefetch(tf.data.experimental.AUTOTUNE)
    return dataset_sample


def is_filter_enabled(filter_config: Bunch) -> bool:
    return filter_config is not None and filter_config.IS_ENABLED


def is_ensemble_evaluation(model_config: Bunch) -> bool:
    return getattr(model_config, 'RUN_IDS', False)


def is_multiartifact_evaluation(data_config: Bunch) -> bool:
    return getattr(data_config, "N_ARTIFACTS", 1) > 1



'''
This file contains a data-generator that the preprocessed data. Note: It does not work on ETL-data.

Note: It currently works on pointcloud-data only!
'''


from __future__ import absolute_import
import os
import logging
import logging.config
import numpy as np
import glob2 as glob
import random
import progressbar
from pyntcloud import PyntCloud
import multiprocessing as mp
import pickle
from . import utils
from bunch import Bunch

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


class PreprocessedDataGenerator(object):
    """
    This class generates data for training.
    """

    def __init__(
        self,
        dataset_path,
        input_type,
        output_targets=["height"],
        filter=None,
        sequence_length=0,
        image_target_shape=(160, 90),
        voxelgrid_target_shape=(32, 32, 32),
        voxel_size_meters=0.01,
        voxelgrid_random_rotation=False,
        pointcloud_target_size=32000,
        pointcloud_subsampling_method="random",
        pointcloud_random_rotation=False,
        rgbmap_target_width=512,
        rgbmap_target_height=512,
        rgbmap_scale_factor=1.5,
        rgbmap_axis="vertical"
    ):
        """
        Initializes a DataGenerator.

        Args:
            dataset_path (string): Where the raw data is.
            input_type (string): Specifies how the input-data for the Neural Network looks like.
                                 Either 'image', 'pointcloud', 'voxgrid'.
            output_targets (list of strings): A list of targets for the Neural Network.
                                              For example *['height', 'weight']*.
            sequence_length (int): Specifies the lenght of the sequences. 0 would yield no sequence at all.
            image_target_shape (2D tuple of ints): Target shape of the images.
            voxelgrid_target_shape (3D tuple of ints): Target shape of the voxelgrids.
            voxel_size_meters (float): Size of the voxels. That is, edge length.
            voxelgrid_random_rotation (bool): If True voxelgrids will be rotated randomly.
            pointcloud_target_size (int): Target size of the pointclouds.
            pointcloud_random_rotation (bool): If True pointclouds will be rotated randomly.

        """

        # Preconditions.
        assert os.path.exists(dataset_path), "dataset_path must exist: " + str(dataset_path)
        assert isinstance(input_type, str), "input_type must be string: " + str(input_type)
        #assert isinstance(output_targets, list), "output_targets must be list: " + str(output_targets)
        if input_type == "image":
            assert len(image_target_shape) == 2, "image_target_shape must be 2-dimensional: " + \
                str(image_target_shape)
        if input_type == "voxelgrid":
            assert len(voxelgrid_target_shape) == 3, "voxelgrid_target_shape must be 3-dimensional: " + \
                str(voxelgrid_target_shape)

        # Set the dataset-path.
        if input_type == "image":
            self.dataset_path = os.path.join(dataset_path, "jpg")
        elif input_type == "voxelgrid":
            self.dataset_path = os.path.join(dataset_path, "pcd")
        elif input_type == "pointcloud":
            self.dataset_path = os.path.join(dataset_path, "pcd")
        elif input_type == "fusion":
            self.dataset_path = os.path.join(dataset_path, "ply")
        elif input_type == "rgbmap":
            self.dataset_path = os.path.join(dataset_path, "pcd")
        else:
            raise Exception("Unknown input_type: " + self.input_type)

        # Assign the instance-variables.
        self.input_type = input_type
        self.output_targets = output_targets
        self.filter = filter
        self.sequence_length = sequence_length
        self.image_target_shape = image_target_shape
        self.voxelgrid_target_shape = voxelgrid_target_shape
        self.voxel_size_meters = voxel_size_meters
        self.voxelgrid_random_rotation = voxelgrid_random_rotation
        self.pointcloud_target_size = pointcloud_target_size
        self.pointcloud_subsampling_method = pointcloud_subsampling_method
        self.pointcloud_random_rotation = pointcloud_random_rotation
        self.rgbmap_target_width = rgbmap_target_width
        self.rgbmap_target_height = rgbmap_target_height
        self.rgbmap_scale_factor = rgbmap_scale_factor
        self.rgbmap_axis = rgbmap_axis

        # Find all QR-codes.
        self._find_qrcodes()
        assert self.qrcodes != [], "No QR-codes found!"

        # Prepare the data.
        self._prepare_qrcodes_dictionary()

    def _find_qrcodes(self):
        """
        Finds all QR-codes.

        Each individual is represented via a unique QR-code. This method extracts the set of QR-codes.
        """

        # Retrieve the QR-codes from the folders.
        paths = glob.glob(os.path.join(self.dataset_path, "*"))
        paths = [path for path in paths if os.path.isdir(path)]
        self.qrcodes = sorted([path.split("/")[-1] for path in paths])

    def _prepare_qrcodes_dictionary(self):

        self.all_pcd_paths = []
        self.all_jpg_paths = []

        self.qrcodes_dictionary = {}
        for qrcode in self.qrcodes:
            # Getting all files that belong to the QR-code.
            glob_search_path = os.path.join(self.dataset_path, qrcode)
            preprocessed_paths = glob.glob(os.path.join(glob_search_path, "*.p"))

            assert len(preprocessed_paths) != 0, "ERROR: No files found at {}!".format(
                glob_search_path)

            # Filter the paths if specified.
            if self.filter is not None:
                if self.filter == "front":
                    filter_for = "104"
                elif self.filter == "360":
                    filter_for = "107"
                elif self.filter == "back":
                    filter_for = "110"
                preprocessed_paths = [path for path in preprocessed_paths if os.path.basename(
                    path).split("_")[-2] == filter_for]

            # Done.
            self.qrcodes_dictionary[qrcode] = preprocessed_paths

    def analyze_files(self):

        for qrcode in self.qrcodes:
            logger.info("QR-code: %s", qrcode)
            logger.info("  Number of samples: %d", len(self.qrcodes_dictionary[qrcode]))

    def generate(self, size, qrcodes_to_use=None, verbose=False, workers=1):

        if qrcodes_to_use is None:
            qrcodes_to_use = self.qrcodes

        logger.info("Using %d workers...", workers)

        # Main loop for single processing.
        if workers == 1:
            while True:
                yield generate_data(self, size, qrcodes_to_use, verbose, None)

        # Main loop for multi processing.
        elif workers > 1:
            pool = mp.Pool(workers)
            self.pool = pool

            self_bunch = Bunch()
            self_bunch.qrcodes_dictionary = self.qrcodes_dictionary
            self_bunch.sequence_length = self.sequence_length
            self_bunch.output_targets = self.output_targets
            self_bunch.input_type = self.input_type
            self_bunch.pointcloud_target_size = self.pointcloud_target_size
            self_bunch.pointcloud_subsampling_method = self.pointcloud_subsampling_method

            self_bunch = mp.Manager().dict(self_bunch)

            # Create chunks of almost equal size.
            subset_sizes = [0] * workers
            subset_sizes[0:workers - 1] = [size // workers] * (workers - 1)
            subset_sizes[workers - 1] = size - sum(subset_sizes[0:workers - 1])
            subset_sizes = [s for s in subset_sizes if s > 0]
            assert sum(subset_sizes) == size

            while True:

                # Spawn a couple of workers and get the results.
                process_target = generate_data
                multiple_results = [pool.apply_async(
                    process_target,
                    (self_bunch, subset_size, qrcodes_to_use, verbose, "return_values")
                ) for subset_size in subset_sizes]

                return_values_list = [res.get(timeout=1) for res in multiple_results]

                # Merge data.
                x_inputs_arrays = []
                y_outputs_arrays = []
                for return_values in return_values_list:
                    x_inputs_arrays.append(return_values[0])
                    y_outputs_arrays.append(return_values[1])
                x_inputs = np.concatenate(x_inputs_arrays)
                y_outputs = np.concatenate(y_outputs_arrays)

                # Done.
                yield x_inputs, y_outputs

            else:
                raise Exception("Unexpected value for 'workers' " + str(workers))

    def _create_voxelgrid_from_pointcloud(self, pointcloud, augmentation=True):
        if self.voxelgrid_random_rotation is True and augmentation is True:
            pointcloud = self._rotate_point_cloud(pointcloud)

        # Create voxelgrid from pointcloud.
        pointcloud = PyntCloud(pointcloud)
        voxelgrid_id = pointcloud.add_structure(
            "voxelgrid", size_x=self.voxel_size_meters, size_y=self.voxel_size_meters, size_z=self.voxel_size_meters)
        voxelgrid = pointcloud.structures[voxelgrid_id].get_feature_vector(mode="density")

        # Do the preprocessing.
        # if preprocess is True:
        #     voxelgrid = utils.ensure_voxelgrid_shape(voxelgrid, self.voxelgrid_target_shape)
        #     assert voxelgrid.shape == self.voxelgrid_target_shape

        return voxelgrid

    def _rotate_point_cloud(self, point_cloud):

        rotation_angle = np.random.uniform() * 2 * np.pi
        cosval = np.cos(rotation_angle)
        sinval = np.sin(rotation_angle)
        rotation_matrix = np.array([[cosval, sinval, 0],
                                    [-sinval, cosval, 0],
                                    [0, 0, 1]])

        rotated_data = np.zeros(point_cloud.shape, dtype=np.float32)
        for k in range(point_cloud.shape[0]):

            shape_pc = point_cloud[k, ...]
            rotated_data[k, ...] = np.dot(shape_pc.reshape((-1, 3)), rotation_matrix)

        return rotated_data

    def finish(self):
        if hasattr(self, "pool"):
            self.pool.terminate()
            self.pool.join()


def generate_data(class_self, size, qrcodes_to_use, verbose, return_control="return_values"):

    if isinstance(class_self, type(Bunch)) is False:
        class_self = Bunch(dict(class_self))

    if verbose is True:
        logger.info("Generating QR-codes to be used: %s", qrcodes_to_use)

    assert size != 0

    x_inputs = []
    y_outputs = []

    if verbose is True:
        bar = progressbar.ProgressBar(max_value=size)
    while len(x_inputs) < size:

        # Get a random QR-code.
        qrcode = random.choice(qrcodes_to_use)

        # Get targets and paths randomly.
        if qrcode not in class_self.qrcodes_dictionary.keys():
            continue

        # Get a sample.
        x_input = None
        y_output = None

        # Get the input. Not dealing with sequences.
        if class_self.sequence_length == 0:
            if len(class_self.qrcodes_dictionary[qrcode]) > 0:
                preprocessed_path = random.choice(class_self.qrcodes_dictionary[qrcode])
                with open(preprocessed_path, "rb") as file:
                    (pointcloud, targets) = load_pointcloud_and_target(
                        file, class_self.output_targets)
                    assert pointcloud.shape[0] != 0, "Empty pointcloud in file {}.".format(
                        preprocessed_path)

                x_input = get_input(class_self, pointcloud)
                y_output = targets
        """
        # Get the input. Dealing with sequences here.
        else:
            preprocessed_paths = np.array(class_self.qrcodes_dictionary[qrcode])

            # Do not touch the QR-code if it does not have enough samples.
            if len(preprocessed_paths) < class_self.sequence_length:
                continue

            # Get some random indices that are in order.
            indices = np.arange(len(preprocessed_paths))
            np.random.shuffle(indices)
            indices = indices[:class_self.sequence_length]
            indices = np.sort(indices)

            preprocessed_paths = preprocessed_paths[indices]
            x_input, file_path = [], []
            for preprocessed_path in preprocessed_paths:
                with open(preprocessed_path, "rb") as file:
                    (pointcloud, targets) = load_pointcloud_and_target(file, class_self.output_targets)
                    assert pointcloud.shape[0] != 0, "Empty pointcloud in file {}.".format(preprocessed_path)
                    try:
                        x_input.append(get_input(class_self, pointcloud))
                    except:
                        print(pointcloud.shape, preprocessed_path)
                        exit(0)
                    file_path.append(preprocessed_path)

            x_input = np.array(x_input)
            y_output = targets
        """

        # Got a proper sample.
        if x_input is not None and y_output is not None:
            x_inputs.append(x_input)
            y_outputs.append(y_output)

        assert len(x_inputs) == len(y_outputs)

        if verbose is True:
            bar.update(len(x_inputs))

    if verbose is True:
        bar.finish()

    assert len(x_inputs) == size
    assert len(y_outputs) == size

    # Turn everything into ndarrays.
    x_inputs = np.array(x_inputs)
    y_outputs = np.array(y_outputs)

    # Prepare result values.
    assert len(x_inputs) == size
    assert len(y_outputs) == size
    return_values = (x_inputs, y_outputs)

    # This is used in multiprocessing. Creates a pickle file and puts the data there.
    def pickle_data(data):
        pickle_path = utils.get_datetime_string() + str(random.random()) + ".temp"
        pickle.dump(data, open(pickle_path, "wb"))
        return pickle_path

    # Just return the path to the pickle file.

    # Return the values directly.
    if return_control == "return_values":
        return return_values

    # Return path to pickled values.
    elif return_control == "return_pickle_path":
        return pickle_data(return_values)

    # Pickle values and store the path in queue.
    elif isinstance(return_control, type(mp.Queue)):
        output_queue = return_control
        output_queue.put(pickle_data(return_values))

    # Should not happen.
    else:
        raise Exception("Unexpected {}".format(return_control))


def load_pointcloud_and_target(file, output_targets):
    (pointcloud, targets) = pickle.load(file)
    if output_targets == ["height"]:
        targets = targets[0:1]
    elif output_targets == ["weight"]:
        targets = targets[1:]
    return (pointcloud, targets)


def get_input(class_self, pointcloud):

    # Get a random image.
    if class_self.input_type == "image":
        raise Exception("Not expected to work with image-data.")

    # Get a random voxelgrid.
    elif class_self.input_type == "voxelgrid":
        voxelgrid = class_self._create_voxelgrid_from_pointcloud(pointcloud)
        x_input = voxelgrid

    # Get a fused point cloud.
    elif class_self.input_type == "fusion":
        pointcloud = utils.subsample_pointcloud(
            pointcloud, class_self.pointcloud_target_size, class_self.pointcloud_subsampling_method, list(range(7)))
        x_input = pointcloud

    # Get a random pointcloud.
    elif class_self.input_type == "pointcloud":
        pointcloud = utils.subsample_pointcloud(
            pointcloud, class_self.pointcloud_target_size, class_self.pointcloud_subsampling_method)
        x_input = pointcloud

    # Get a random pointcloud.
    elif class_self.input_type == "rgbmap":
        rgb_map = utils.pointcloud_to_rgb_map(
            pointcloud,
            class_self.rgbmap_target_width,
            class_self.rgbmap_target_height,
            class_self.rgbmap_scale_factor,
            class_self.rgbmap_axis
        )
        x_input = rgb_map

    # Should not happen.
    else:
        raise Exception("Unknown input_type: " + class_self.input_type)

    return x_input


def create_datagenerator_from_parameters(dataset_path, dataset_parameters):
    logger.info("Creating data-generator...")
    datagenerator = PreprocessedDataGenerator(
        dataset_path=dataset_path,
        input_type=dataset_parameters["input_type"],
        output_targets=dataset_parameters["output_targets"],
        filter=dataset_parameters.get("filter", None),
        sequence_length=dataset_parameters.get("sequence_length", 0),
        voxelgrid_target_shape=dataset_parameters.get("voxelgrid_target_shape", None),
        voxel_size_meters=dataset_parameters.get("voxel_size_meters", None),
        voxelgrid_random_rotation=dataset_parameters.get("voxelgrid_random_rotation", None),
        pointcloud_target_size=dataset_parameters.get("pointcloud_target_size", None),
        pointcloud_subsampling_method=dataset_parameters.get("pointcloud_subsampling_method", None),
        pointcloud_random_rotation=dataset_parameters.get("pointcloud_random_rotation", None),
        rgbmap_target_width=dataset_parameters.get("rgbmap_target_width", None),
        rgbmap_target_height=dataset_parameters.get("rgbmap_target_height", None),
        rgbmap_scale_factor=dataset_parameters.get("rgbmap_scale_factor", None),
        rgbmap_axis=dataset_parameters.get("rgbmap_axis", None),
    )
    #datagenerator.print_statistics()
    return datagenerator


def get_dataset_path(root_path="/whhdata/preprocessed"):
    if os.path.exists("etldatasetpath.txt"):
        with open("etldatasetpath.txt", "r") as file:
            dataset_path = file.read().replace("\n", "")
    else:
        # Finding the latest.
        dataset_paths = glob.glob(os.path.join(root_path, "*"))
        dataset_paths = [
            dataset_path for dataset_path in dataset_paths if os.path.isdir(dataset_path)]
        dataset_path = list(reversed(sorted(dataset_paths)))[0]

    return dataset_path



import os

from bunch import Bunch

CONFIG_NAME = os.path.splitext(os.path.basename(__file__))[0]

# Details of model used for evaluation
MODEL_CONFIG = Bunch(dict(
    EXPERIMENT_NAME='q3-depthmap-plaincnn-height-95k',
    RUN_ID='q3-depthmap-plaincnn-height-95k_1597988908_42c4ef33',  # Run 3

    INPUT_LOCATION='outputs',
    NAME='best_model.ckpt',
))

FILTER_CONFIG = Bunch(dict(
    IS_ENABLED=True,  # 'False'
    EXPERIMENT_NAME='q4-rgb-plaincnn-classifaction-standing-lying-8k',

    RUN_ID='q4-rgb-plaincnn-classifaction-standing-lying-8k_1602316038_3ebdb326',  # Run 3
    # RUN_ID = 'q3-depthmap-plaincnn-height-95k_1600451633_cb44f6db',     #Run 17

    INPUT_LOCATION='outputs',
    NAME='best_filter.h5',
))

EVAL_CONFIG = Bunch(dict(
    # Name of evaluation
    NAME='q3-depthmap-plaincnn-height-95k_run_03',

    # Experiment in Azure ML which will be used for evaluation
    EXPERIMENT_NAME="QA-pipeline",
    CLUSTER_NAME="gpu-cluster",

    # Used for Debug the QA pipeline
    DEBUG_RUN=False,

    # Will run eval on specified # of scan instead of full dataset
    DEBUG_NUMBER_OF_SCAN=5,

    SPLIT_SEED=0,
))

# Details of Evaluation Dataset
DATA_CONFIG = Bunch(dict(
    # Name of evaluation dataset
    NAME='anon-depthmap-rgb-timestamp',
    IMAGE_TARGET_HEIGHT=240,
    IMAGE_TARGET_WIDTH=180,
    # Batch size for evaluation
    BATCH_SIZE=512,
    NORMALIZATION_VALUE=7.5,
    # Parameters for dataset generation.

    # 0 is height, 1 is weight, 2 is muac, 3 is weight,
    # 4 is sex('male' or 'female'), 5 is quality ('good' or 'bad'), 6 is test
    TARGET_INDEXES=[0, 3, 4, 5],

    CODES=['100', '101', '102', '200', '201', '202']
))

# Result configuration for result generation after evaluation is done
RESULT_CONFIG = Bunch(dict(
    # Error margin on various ranges
    #EVALUATION_ACCURACIES = [.2, .4, .8, 1.2, 2., 2.5, 3., 4., 5., 6.]
    ACCURACIES=[.2, .4, .6, 1., 1.2, 2., 2.5, 3., 4., 5., 6.],  # 0.2cm, 0.4cm, 0.6cm, 1cm, ...
    ACCURACY_MAIN_THRESH=1.0,  # 1cm
    AGE_BUCKETS=[0, 1, 2, 3, 4, 5],

    COLUMNS=['qrcode', 'artifact', 'scantype', 'GT', 'predicted'],

    # uncertainty
    USE_UNCERTAINTY=False,  # Flag to enable model uncertainty calculation

    # path of csv file in the experiment which final result is stored
    SAVE_PATH=f'./outputs/{CONFIG_NAME}',
))



import argparse
import logging
import logging.config
from azureml.core import Workspace
from azureml.core.authentication import AuthenticationException, AzureCliAuthentication

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def get_auth():
    """Authentication to access workspace"""
    try:
        auth = AzureCliAuthentication()
        auth.get_authentication_header()
    except AuthenticationException:
        logger.info("Authentication Error Occured")

    return auth


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-sid", "--subscription_id", help="Subscription ID")
    parser.add_argument("-rg", "--resource_group", help="Resource Group")
    parser.add_argument("-wn", "--workspace_name", help="Workspace  Name")

    args = parser.parse_args()

    workspace = Workspace(subscription_id=args.subscription_id,
                          resource_group=args.resource_group,
                          workspace_name=args.workspace_name,
                          auth=get_auth())

    logger.info("Workspace Details")
    logger.info(workspace.get_details())

    logger.info("Success of Authentication and Workspace Setup")

    workspace.write_config()
    logger.info("Saved config file")



NAMES = [
    "__background__",
    "person",
    "bicycle",
    "car",
    "motorcycle",
    "airplane",
    "bus",
    "train",
    "truck",
    "boat",
    "traffic light",
    "fire hydrant",
    "N/A",
    "stop sign",
    "parking meter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "N/A",
    "backpack",
    "umbrella",
    "N/A",
    "N/A",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sports ball",
    "kite",
    "baseball bat",
    "baseball glove",
    "skateboard",
    "surfboard",
    "tennis racket",
    "bottle",
    "N/A",
    "wine glass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hot dog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "couch",
    "potted plant",
    "bed",
    "N/A",
    "dining table",
    "N/A",
    "N/A",
    "toilet",
    "N/A",
    "tv",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cell phone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "N/A",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddy bear",
    "hair drier",
    "toothbrush",
]



"""Preprocessing utilities

In order to preprocess ZIP file to extract a depthmap, we use this code:
https://github.com/Welthungerhilfe/cgm-rg/blob/92efa0febb91c9656ce8e5dbfad953ff7ce721a9/src/utils/preprocessing.py#L12

file of minor importance:
https://github.com/Welthungerhilfe/cgm-ml/blob/c8be9138e025845bedbe7cfc0d131ef668e01d4b/old
/cgm_database/command_preprocess.py#L92
"""

from pathlib import Path
import pickle
from typing import List, Tuple

import numpy as np
from skimage.transform import resize

from cgmml.common.depthmap_toolkit.depthmap import Depthmap

TOOLKIT_DIR = Path(__file__).parents[1] / 'depthmap_toolkit'
CALIBRATION_FPATH = TOOLKIT_DIR / "camera_calibration_p30pro_EU.txt"
NORMALIZATION_VALUE = 7.5
IMAGE_TARGET_HEIGHT, IMAGE_TARGET_WIDTH = 180, 240


def preprocess_depthmap(depthmap: np.ndarray) -> np.ndarray:
    return depthmap.astype("float32")


def preprocess(depthmap: np.ndarray) -> np.ndarray:
    depthmap = preprocess_depthmap(depthmap)
    depthmap = depthmap / NORMALIZATION_VALUE

    width, height = depthmap.shape
    assert width / 1 == height / IMAGE_TARGET_HEIGHT, f"\
        {width} / {IMAGE_TARGET_WIDTH} == {height} / {IMAGE_TARGET_HEIGHT}"
    depthmap = resize(depthmap, (IMAGE_TARGET_WIDTH, IMAGE_TARGET_HEIGHT))

    depthmap = depthmap.reshape((depthmap.shape[0], depthmap.shape[1], 1))
    return depthmap


def preprocess_rgb(rgb_array: np.ndarray) -> np.ndarray:
    """Preprocess RGB

    Args:
        rgb_array: shape (height, width, 3)

    Returns:
        proprocessed rgb array
    """
    rgb_array = rgb_array.astype("float32")
    rgb_array = rgb_array / 255.

    width, height, _ = rgb_array.shape
    assert width / IMAGE_TARGET_WIDTH == height / IMAGE_TARGET_HEIGHT, f"\
        {width} / {IMAGE_TARGET_WIDTH} == {height} / {IMAGE_TARGET_HEIGHT}"
    rgb_array = resize(rgb_array, (IMAGE_TARGET_WIDTH, IMAGE_TARGET_HEIGHT, 3))

    return rgb_array


def create_layers(depthmap_fpath: str) -> Tuple[np.ndarray, dict]:
    dmap = Depthmap.create_from_zip_absolute(depthmap_fpath, rgb_fpath=None, calibration_fpath=CALIBRATION_FPATH)
    depthmap = dmap.depthmap_arr  # shape: (width, height)
    depthmap = preprocess(depthmap)
    layers = depthmap
    metadata = {
        'device_pose': dmap.device_pose,
        'raw_header': dmap.header,
    }
    return layers, metadata


def create_layers_rgbd(depthmap_fpath: str, rgb_fpath: str, should_rotate_rgb) -> Tuple[np.ndarray, dict]:
    dmap = Depthmap.create_from_zip_absolute(depthmap_fpath, rgb_fpath, CALIBRATION_FPATH, should_rotate_rgb)

    depthmap = dmap.depthmap_arr  # shape: (longer, shorter)
    depthmap = preprocess(depthmap)  # shape (longer, shorter, 1)

    rgb = dmap.rgb_array  # shape (longer, shorter, 3)
    rgb = preprocess_rgb(rgb)  # shape (longer, shorter, 3)

    layers = np.concatenate([
        depthmap,  # shape (longer, shorter, 1)
        rgb,  # shape (longer, shorter, 3)
    ], axis=2)  # shape (longer, shorter, 4)

    metadata = {
        'device_pose': dmap.device_pose,
        'raw_header': dmap.header,
    }
    return layers, metadata


def create_layers_from_multiple_paths(fpaths: List[str]) -> np.ndarray:
    depthmaps = []
    for fpath in fpaths:
        depthmap = create_layers(fpath)
        depthmaps.append(depthmap)
    depthmaps = np.array(depthmaps)
    return depthmaps


class ArtifactProcessor:
    def __init__(self, input_dir: str, output_dir: str, dataset_type: str, should_rotate_rgb: bool = False):
        self.input_dir = input_dir
        self.output_dir = output_dir
        assert dataset_type in ['depthmap', 'rgbd']
        self.dataset_type = dataset_type
        self.should_rotate_rgb = should_rotate_rgb

    def create_and_save_pickle(self, artifact_dict: dict) -> str:
        """Side effect: Saves and returns file path"""
        # Prepare data to save
        zip_input_full_path = f"{self.input_dir}/{artifact_dict['file_path']}"

        if self.dataset_type == 'depthmap':
            layers, metadata = create_layers(zip_input_full_path)
        elif self.dataset_type == 'rgbd':
            rgb_input_full_path = f"{self.input_dir}/{artifact_dict['file_path_rgb']}"
            layers, metadata = create_layers_rgbd(zip_input_full_path, rgb_input_full_path, self.should_rotate_rgb)
        else:
            raise NameError(self.dataset_type)
        target_dict = {**artifact_dict, **metadata}

        # Prepare path
        timestamp = artifact_dict['timestamp']
        scan_id = artifact_dict['scan_id']
        scan_step = artifact_dict['scan_step']
        order_number = artifact_dict['order_number']
        pickle_output_path = f"scans/{scan_id}/{scan_step}/pc_{scan_id}_{timestamp}_{scan_step}_{order_number}.p"

        # Write into pickle
        pickle_output_full_path = f"{self.output_dir}/{pickle_output_path}"
        Path(pickle_output_full_path).parent.mkdir(parents=True, exist_ok=True)
        pickle.dump((layers, target_dict), open(pickle_output_full_path, "wb"))

        return pickle_output_full_path



from pathlib import Path

from azureml.core import Environment, Workspace


def cgm_environment(workspace: Workspace,
                    curated_env_name: str,
                    env_exist: bool,
                    fpath_env_yml: Path = None) -> Environment:
    if env_exist:
        return Environment.get(workspace=workspace, name=curated_env_name)
    if fpath_env_yml is None:
        from cgmml.common.endpoints.constants import REPO_DIR  # noqa: E402
        fpath_env_yml = REPO_DIR / "environment_train.yml"
    else:
        fpath_env_yml = Path(fpath_env_yml)
    cgm_env = Environment.from_conda_specification(name=curated_env_name, file_path=fpath_env_yml)
    cgm_env.docker.enabled = True
    cgm_env.docker.base_image = 'mcr.microsoft.com/azureml/openmpi4.1.0-cuda11.0.3-cudnn8-ubuntu18.04'
    return cgm_env



import logging
import logging.config

import numpy as np

from cgmml.common.depthmap_toolkit.constants import MASK_CHILD
from cgmml.common.depthmap_toolkit.depthmap import Depthmap
from cgmml.common.depthmap_toolkit.depthmap_utils import calculate_boundary, get_smoothed_pixel

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

CHILD_HEAD_HEIGHT_IN_METERS = 0.25
PATTERN_LENGTH_IN_METERS = 0.2
IDX_RED = 0
IDX_GREEN = 1
IDX_BLUE = 2


def blur_face(data: np.ndarray, highest_point: np.ndarray, dmap: Depthmap) -> np.ndarray:
    """Faceblur of the detected standing child.

    It uses the highest point of the child and blur all pixels in distance less than CHILD_HEAD_HEIGHT_IN_METERS.

    Args:
        data: existing canvas to blur
        highest_point: 3D point. The surroundings of this point will be blurred.
        dmap: depthmap

    Returns:
        Canvas like data with face blurred.
    """
    output = np.copy(data)
    points_3d_arr = dmap.convert_2d_to_3d_oriented()

    # blur RGB data around face
    for x in range(dmap.width):
        for y in range(dmap.height):

            # count distance from the highest child point
            depth = dmap.depthmap_arr[x, y]
            if not depth:
                continue
            point = points_3d_arr[:, x, y]

            vector = point - highest_point
            distance = abs(vector[0]) + abs(vector[1]) + abs(vector[2])
            if distance >= CHILD_HEAD_HEIGHT_IN_METERS:
                continue

            # Gausian blur
            index = dmap.height - y - 1
            output[x, index] = get_smoothed_pixel(data, x, index, 5)

    return output


def draw_boundary(output: np.ndarray, aabb: np.array, color: np.array):
    height = output.shape[1]
    x1 = aabb[0]
    x2 = aabb[2]
    y1 = height - aabb[3] - 1
    y2 = height - aabb[1] - 1

    output[x1:x2, y1, :] = color
    output[x1:x2, y2, :] = color
    output[x1, y1:y2, :] = color
    output[x2, y1:y2, :] = color


def render_confidence(dmap: Depthmap):
    confidence = dmap.confidence_arr
    confidence[confidence == 0.] = 1.

    confidence = np.fliplr(confidence)  # flip left-right
    output = np.stack([confidence, confidence, confidence], axis=2)
    return output


def render_depth(dmap: Depthmap, use_smooth=False) -> np.ndarray:
    """Render depthmap into a 2D image.

    We assume here that all values in dmap.depthmap_arr are positive.

    A distance of 0m is visualized in white.
    A distance of 2m is visualized in black.

    flashlight analogy: close-to-cam data is white
    """
    if use_smooth:
        dmap_arr = np.minimum(dmap.depthmap_arr_smooth / 2., 1.)
    else:
        dmap_arr = np.minimum(dmap.depthmap_arr / 2., 1.)

    cond = (dmap_arr != 0.)
    dmap_arr[cond] = 1. - dmap_arr[cond]

    dmap_arr = np.fliplr(dmap_arr)  # flip left-right
    output = np.stack([dmap_arr, dmap_arr, dmap_arr], axis=2)
    return output


def render_normal(dmap: Depthmap) -> np.ndarray:
    """Render normal vectors

    How normal vector are visualized:
    When a vector has (x,y,z)=(1,0,0), this will show in red color.
    When a vector has (x,y,z)=(0,1,0), this will show in green color (e.g. floor).
    When a vector has (x,y,z)=(0,0,1), this will show in blue color.
    """

    points_3d_arr = dmap.convert_2d_to_3d_oriented(should_smooth=True)
    normal = dmap.calculate_normalmap_array(points_3d_arr)

    # We can't see negative values, so we take the absolute value
    normal = abs(normal)  # shape: (3, width, height)

    output = np.moveaxis(normal, 0, -1)
    output = np.fliplr(output)  # flip left-right
    return output


def render_rgb(dmap: Depthmap) -> np.ndarray:
    output = np.copy(dmap.rgb_array)  # shape (width, width, 3)
    output = output / 255.
    return output


def render_segmentation(floor: float,
                        mask: np.ndarray,
                        dmap: Depthmap) -> np.ndarray:
    output = np.zeros((dmap.width, dmap.height, 3))

    point = dmap.convert_2d_to_3d_oriented(should_smooth=True)
    normal = dmap.calculate_normalmap_array(point)

    for x in range(dmap.width):
        for y in range(dmap.height):

            # get depth value
            depth = dmap.depthmap_arr[x, y]
            if not depth:
                continue

            # segmentation visualisation
            fog = depth * depth
            elevation = point[1, x, y] - floor
            horizontal = (elevation % PATTERN_LENGTH_IN_METERS) / PATTERN_LENGTH_IN_METERS / fog
            vertical_x = (point[0, x, y] % PATTERN_LENGTH_IN_METERS) / PATTERN_LENGTH_IN_METERS / fog
            vertical_z = (point[2, x, y] % PATTERN_LENGTH_IN_METERS) / PATTERN_LENGTH_IN_METERS / fog
            vertical = (vertical_x + vertical_z) / 2.0 / fog
            index = dmap.height - y - 1

            if mask[x, y] == MASK_CHILD:
                # Red + Green = Yellow
                output[x, index, IDX_RED] = horizontal
                output[x, index, IDX_GREEN] = horizontal
            elif abs(normal[1, x, y]) < 0.5:
                output[x, index, IDX_RED] = horizontal
            else:
                if abs(point[1, x, y] - floor) < 0.1:
                    output[x, index, IDX_BLUE] = vertical
                else:
                    output[x, index, IDX_GREEN] = vertical

    # Ensure pixel clipping
    np.clip(output, 0., 1., output)

    # Show the boundary of the child
    color = [1, 0, 1]  # purple
    if dmap.is_child_fully_visible(mask):
        color = [0, 1, 0]  # green
    aabb = calculate_boundary(mask == MASK_CHILD)
    draw_boundary(output, aabb, color)

    return output


def render_plot(dmap: Depthmap) -> np.ndarray:
    # detect floor and child
    floor: float = dmap.get_floor_level()
    mask = dmap.segment_child(floor)  # dmap.detect_floor(floor)
    highest_point: np.ndarray = dmap.get_highest_point(mask)

    # prepare plots
    output_plots = [
        render_depth(dmap),
        render_normal(dmap),
        render_segmentation(floor, mask, dmap),
        render_confidence(dmap),
    ]
    if dmap.has_rgb:
        output_rgb = render_rgb(dmap)
        output_rgb = blur_face(output_rgb, highest_point, dmap)
        output_plots.append(output_rgb)

    output = np.concatenate(output_plots, axis=1)
    logger.info('height=%fm', highest_point[1] - floor)
    return output



import json

import numpy as np
from tensorflow.keras.models import load_model

MODEL = None


def init():
    global MODEL
    MODEL = load_model(
        '/var/azureml-app/azureml-models/2021q1-depthmap-ensemble-height-95k/1/outputs/best_model.ckpt/', compile=False)


def run(input: json) -> list:
    data_json = json.loads(input)
    data_np = np.array(data_json['data'])
    result = MODEL.predict(data_np)
    return result.tolist()



from setuptools import setup, find_packages

setup(
    name='cgmzscore',
    version='3.0.0',

    license='GPL',

    author='Prajwal Kumar Singh',
    author_email='prajwalsingh651@gmail.com',

    packages=find_packages(),
    include_package_data=True,

    description="z-scores of anthropometric measurements of children below 5 years  based on WHO",
    long_description=open('README.rst').read(),

    classifiers=[
        'Intended Audience :: Healthcare Industry',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Operating System :: OS Independent',
    ]
)



from bunch import Bunch

from constants import REPO_DIR

CONFIG = Bunch(dict(
    MODEL_NAME="2021q1-depthmap-ensemble-height-95k",
    VERSION='1',
    ENDPOINT_NAME='aci-21q1-ensemble-h-95k-m01-v1',
    EXPERIMENT_NAME='2021q1-depthmap-ensemble-height-95k',
    RUN_ID='2021q1-depthmap-ensemble-height-95k_1622230334_67c64a77',
    LOCALTEST=False,
    TEST_FILES=[REPO_DIR / 'cgmml/common/depthmap_toolkit/tests/static_files/4ed427b5-3fd9-4f4d-8e58-19e39c7d77b6',
                REPO_DIR / 'cgmml/common/depthmap_toolkit/tests/static_files/4ed427b5-3fd9-4f4d-8e58-19e39c7d77b6']
))



# Child Growth Monitor - Free Software for Zero Hunger
# Copyright (c) 2019 Dr. Christian Pfitzner <christian.pfitzner@th-nuernberg.de> for Welthungerhilfe
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
import xmltodict

#calib_file = r'/whhdata/calibration.xml' # todo make this a parameter


def get_intrinsic_matrix(calib_file):
    with open(calib_file) as fd:
        calib = xmltodict.parse(fd.read())

    arr = calib['rig']['camera'][1]['camera_model']['params']
    fu, fv, u0, v0, k1, k2, k3 = np.fromstring(arr.replace('[', '').replace(
        ']', ''),
        sep=';')
    _gamma = 1
    intrinsic = np.array([[fu, _gamma, u0, 0], [0, fv, v0, 0], [0, 0, 1, 0]])
    return intrinsic


def get_intrinsic_matrix_depth(calib_file):
    with open(calib_file) as fd:
        calib = xmltodict.parse(fd.read())

    arr = calib['rig']['camera'][0]['camera_model']['params']
    fu, fv, u0, v0, k1, k2, p1, p2, k3 = np.fromstring(arr.replace(
        '[', '').replace(']', ''),
        sep=';')
    _gamma = 1
    intrinsic = np.array([[fu, _gamma, u0, 0], [0, fv, v0, 0], [0, 0, 1, 0]])
    return intrinsic


def get_k(calib_file):
    with open(calib_file) as fd:
        calib = xmltodict.parse(fd.read())

    arr = calib['rig']['camera'][1]['camera_model']['params']
    fu, fv, u0, v0, k1, k2, k3 = np.fromstring(arr.replace('[', '').replace(
        ']', ''),
        sep=';')
    np.array([[fu, 0, u0, 0], [0, fv, v0, 0], [0, 0, 1, 0]])
    return k1, k2, k3


def get_k_depth(calib_file):
    with open(calib_file) as fd:
        calib = xmltodict.parse(fd.read())

    arr = calib['rig']['camera'][0]['camera_model']['params']
    fu, fv, u0, v0, k1, k2, p1, p2, k3 = np.fromstring(arr.replace(
        '[', '').replace(']', ''),
        sep=';')
    np.array([[fu, 0, u0, 0], [0, fv, v0, 0], [0, 0, 1, 0]])
    return k1, k2, k3


def get_extrinsic_matrix(calib_file, idx=1):
    with open(calib_file) as fd:
        calib = xmltodict.parse(fd.read())

    arr = calib['rig']['extrinsic_calibration'][idx]['A_T_B']
    arr = arr.split(';')
    arr = [x.replace('[', '').replace(']', '') for x in arr]
    mat = np.array([np.fromstring(x, sep=',') for x in arr])
    mat[:3, :3] = mat[:3, :3].T  # maybe transpose?
    mat = np.vstack([mat, np.array([0, 0, 0, 1])])
    return mat


def get_extrinsic_matrix_depth(calib_file, idx=1):
    with open(calib_file) as fd:
        calib = xmltodict.parse(fd.read())

    arr = calib['rig']['extrinsic_calibration'][0]['A_T_B']
    arr = arr.split(';')
    arr = [x.replace('[', '').replace(']', '') for x in arr]
    mat = np.array([np.fromstring(x, sep=',') for x in arr])
    mat[:3, :3] = mat[:3, :3].T  # maybe transpose?
    mat = np.vstack([mat, np.array([0, 0, 0, 1])])
    return mat



import logging
import logging.config

from cgmml.common.rgbd_toolkit.cgm_fusion import utility

# flake8: noqa: E501

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)

#from cgm_fusion import calibration


# import glob, os
# os.chdir("/localssd/qrcode/")
# for file in glob.glob("*.ply"):
#     logger.info(file)

# import dbutils
# import config


# # get the number of rgb artifacts
# select_sql_statement = "SELECT path FROM artifact WHERE type='pcd';"
# pcd_paths = db_connector.execute(select_sql_statement, fetch_all=True)[0][0]
# logger.info(pcd_paths)

# fusion.get_depth_image_from_point_cloud(calibration_file="dummy", pcd_file="/tmp/cloud_debug.ply", output_file="dummy")
# utility.get_depth_channel(ply_path="/tmp/cloud_debug.ply", output_path_np = "/tmp/output.npy", output_path_png="/tmp/output.png")
# utility.get_rgbd_channel(ply_path="/tmp/cloud_debug.ply", output_path_np = "/tmp/output.npy")

utility.get_all_channel(ply_path="/tmp/cloud_debug.ply",
                        output_path_np="/tmp/output.npy")

# utility.get_viz_channel(ply_path="/tmp/cloud_debug.ply",  channel=4, output_path="/tmp/red.png")



import matplotlib.pyplot as plt
from pathlib import Path
import shutil
import sys

from csv_utils import read_csv
from cgmml.common.depthmap_toolkit.depthmap import Depthmap
from cgmml.common.depthmap_toolkit.visualisation import render_plot

METADATA_DEPTHMAP = 3

REPO_DIR = Path(__file__).parents[4]
EXPORT_DIR = REPO_DIR / 'data' / 'render'

if __name__ == "__main__":

    if len(sys.argv) != 3:
        print('You did not enter raw data path and metadata file path')
        print('E.g.: python evaluation.py rawdata_dir metadata_path')
        sys.exit(1)

    calibration_file = '../../depthmap_toolkit/camera_calibration_p30pro_EU.txt'
    path = sys.argv[1]
    metadata_file = sys.argv[2]

    # Load metadata
    indata = read_csv(metadata_file)
    size = len(indata)

    # Re-create export folder
    try:
        shutil.rmtree(EXPORT_DIR)
    except BaseException:
        print('no previous data to delete')
    EXPORT_DIR.mkdir(parents=True, exist_ok=True)

    # For every depthmap show visualisation
    output = []
    processed = {}
    for index in range(size):
        data = indata[index]

        # Load data
        depthmap_file = path + data[METADATA_DEPTHMAP]
        depthmap_file = depthmap_file.replace('"', '')
        dmap = Depthmap.create_from_zip_absolute(depthmap_file, 0, calibration_file)

        # Render data
        file = str(EXPORT_DIR) + '/' + str(index + 1) + '.png'
        print(file)
        plt.imsave(file, render_plot(dmap))



import logging
import logging.config
import numpy as np
import sys

from csv_utils import read_csv, write_csv
from cgmml.common.depthmap_toolkit.depthmap import Depthmap, is_google_tango_resolution

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
logger.addHandler(handler)

ERROR_THRESHOLDS = [0.2, 0.4, 0.6, 1.0, 1.2, 1.4, 2.0, 2.5, 3.0, 4.0, 5.0, 6.0]
MINIMUM_CHILD_HEIGHT_IN_CM = 45

METADATA_SCAN_ID = 0
METADATA_ORDER = 1
METADATA_ARTIFACT_ID = 2
METADATA_DEPTHMAP = 3
METADATA_RGB = 4
METADATA_MANUAL_HEIGHT = 5
METADATA_MANUAL_WEIGHT = 6
METADATA_MANUAL_MUAC = 7
METADATA_SCAN_VERSION = 8
METADATA_SCAN_TYPE = 9
METADATA_MANUAL_DATE = 10
METADATA_SCAN_DATE = 11
METADATA_HEIGHT = 12
METADATA_ERROR = 13


def filter_metadata(indata: list) -> list:
    size = len(indata)
    output = []
    for index in range(size):

        # Check if the scan version is correct
        data = indata[index]
        if not data[METADATA_SCAN_VERSION].startswith('v0.9'):
            continue

        # Check if it is a standing child
        if not data[METADATA_SCAN_TYPE].startswith('10'):
            continue

        # Check if it is a first frame of the scan
        if int(data[METADATA_ORDER]) != 1:
            continue

        output.append(data)
    return output


def generate_report(indata: list, info: str) -> list:

    # Generate report format
    output = [
        ['Scan type     ', 0.2, 0.4, 0.6, 1.0, 1.2, 1.4, 2.0, 2.5, 3.0, 4.0, 5.0, 6.0],
        ['Standing front', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ['Standing 360  ', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ['Standing back ', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    ]
    count = ['', 0, 0, 0]

    for row in indata:

        # Get scan type
        line = 0
        scan_type = int(row[METADATA_SCAN_TYPE])
        if scan_type == 100:
            line = 1
        if scan_type == 101:
            line = 2
        if scan_type == 102:
            line = 3
        count[line] += 1

        # Update error categories
        error = float(row[METADATA_ERROR])
        for i, error_thresh in enumerate(ERROR_THRESHOLDS, start=1):
            if error <= error_thresh:
                output[line][i] += 1

    # Show result in percents
    for row in range(1, 4):
        for column in range(1, 13):
            output[row][column] = str(100. * output[row][column] / float(max(1, count[row]))) + '%'
    output.append([info])
    return output


def log_rejection(output: list, data: list, reason: str):
    logger.info(reason)
    data.append(reason)
    output.append(data)


def log_report(data: list):
    output = '\n'
    first_row = True
    for row in data:
        first = True
        line = ''
        for value in row:
            if len(line) > 0:
                line = line + ', '
            if first:
                line = line + str(value)
                first = False
            else:
                number = str(value).replace('%', '')
                if len(number) >= 8:
                    number = number[0: 7]
                while len(number) < 8:
                    number = ' ' + number
                if first_row:
                    number = ' ' + number
                else:
                    number = number + '%'
                line = line + number

        output = output + line + '\n'
        first_row = False
    logger.info(output)


if __name__ == "__main__":

    if len(sys.argv) != 3:
        print('You did not enter raw data path and metadata file name')
        print('E.g.: python evaluation.py rawdata_dir metadata_file')
        sys.exit(1)

    calibration_file = '../../depthmap_toolkit/camera_calibration_p30pro_EU.txt'
    path = sys.argv[1]
    metadata_file = path + '/' + sys.argv[2]

    avg_err = 0
    output = []
    rejections = []
    indata = filter_metadata(read_csv(metadata_file))
    size = len(indata)
    for index in range(size):
        logger.info('Processing %d/%d', index + 1, size)
        data = indata[index]

        # Check if it is captured by a new device
        depthmap_file = path + '/' + data[METADATA_DEPTHMAP]
        depthmap_file = depthmap_file.replace('"', '')
        dmap = Depthmap.create_from_zip_absolute(depthmap_file, 0, calibration_file)
        if is_google_tango_resolution(dmap.width, dmap.height):
            log_rejection(rejections, data, 'Skipping because it is not a new device data')
            continue

        # Check if the child is fully visible
        floor: float = dmap.get_floor_level()
        mask = dmap.segment_child(floor)
        if not dmap.is_child_fully_visible(mask):
            log_rejection(rejections, data, 'Skipping because the child is not fully visible')
            continue

        # Filter heights less than MINIMUM_CHILD_HEIGHT_IN_CM
        highest_point: np.ndarray = dmap.get_highest_point(mask)
        height = (highest_point[1] - floor) * 100.0
        if height < MINIMUM_CHILD_HEIGHT_IN_CM:
            limit = str(MINIMUM_CHILD_HEIGHT_IN_CM) + 'cm'
            log_rejection(rejections, data, 'Skipping because the height is less than ' + limit)
            continue

        # Update output
        error = abs(height - float(indata[index][METADATA_MANUAL_HEIGHT]))
        logger.info('Height=%fcm, error=%fcm', height, error)
        data.append(height)
        data.append(error)
        output.append(data)
        avg_err += error
        info = 'Average error=' + str(avg_err / float(len(output))) + 'cm'
        log_report(generate_report(output, info))

    write_csv('output.csv', output)
    write_csv('rejections.csv', rejections)
    write_csv('report.csv', generate_report(output, info))



import os
import logging
import logging.config
from typing import Iterable

import cv2
import face_recognition
import numpy as np


CODES_FRONT_FACING = ("100", "200")
CODES_BACK_FACING = ("102", "202")
CODES_360 = ("101", "201")

RESIZE_FACTOR = 4

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def blur_faces_in_file(source_path: str, target_path: str) -> bool:
    """Blur image

    Returns:
        bool: True if blurred otherwise False
    """

    # Read the image.
    assert os.path.exists(source_path), f"{source_path} does not exist"
    rgb_image = cv2.imread(source_path)
    image = rgb_image[:, :, ::-1]  # RGB -> BGR for OpenCV

    # The images are provided in 90degrees turned. Here we rotate 90degress to the right.
    image = np.swapaxes(image, 0, 1)

    # Scale image down for faster prediction.
    small_image = cv2.resize(image, (0, 0), fx=1. / RESIZE_FACTOR, fy=1. / RESIZE_FACTOR)

    # Find face locations.
    face_locations = face_recognition.face_locations(small_image, model="cnn")

    # Check if image should be used.
    if not should_image_be_used(source_path, number_of_faces=len(face_locations)):
        logger.warn(f"{len(face_locations)} face locations found and not blurred for path: {source_path}")
        return False

    # Blur the image.
    for top, right, bottom, left in face_locations:
        # Scale back up face locations since the frame we detected in was scaled to 1/4 size
        top *= RESIZE_FACTOR
        right *= RESIZE_FACTOR
        bottom *= RESIZE_FACTOR
        left *= RESIZE_FACTOR

        # Extract the region of the image that contains the face.
        face_image = image[top:bottom, left:right]

        # Blur the face image.
        face_image = cv2.GaussianBlur(face_image, ksize=(99, 99), sigmaX=30)

        # Put the blurred face region back into the frame image.
        image[top:bottom, left:right] = face_image

    # Rotate image back.
    image = np.swapaxes(image, 0, 1)

    # Write image to hard drive.
    rgb_image = image[:, :, ::-1]  # BGR -> RGB for OpenCV
    cv2.imwrite(target_path, rgb_image)

    logger.info(f"{len(face_locations)} face locations found and blurred for path: {source_path}")
    return True


def should_image_be_used(source_path: str, number_of_faces: int) -> bool:
    """Determines if an image should be skipped or not."""
    if does_path_belong_to_codes(source_path, CODES_FRONT_FACING):
        return number_of_faces == 1
    elif does_path_belong_to_codes(source_path, CODES_360):
        return True
    elif does_path_belong_to_codes(source_path, CODES_BACK_FACING):
        return True
    else:
        raise NameError(f"{source_path} does not have a correct code")


def does_path_belong_to_codes(path: str, codes: Iterable) -> bool:
    for code in codes:
        if f"_{code}_" in path:
            return True
    return False



"""https://www.who.int/childgrowth/standards/Chap_7.pdf"""
from decimal import Decimal as D


class Zscore:
    """Class to do calculation of z score"""

    def __init__(self, skew: int, median: int, coff: int, measurement: int):
        self.skew = skew
        self.median = median
        self.coff = coff
        self.measurement = measurement

    def calc_stdev(self, number: int):
        """This is not usual Standard Deviation please visit above PDF attached for clarification"""
        value = (1 + (self.skew * self.coff * number))**(1 / self.skew)
        stdev = self.median * value
        return stdev

    def z_score_measurement(self) -> float:
        """
         Z score
                  [y/M(t)]^L(t) - 1
           Zind =  -----------------
                      S(t)L(t)

                |       Zind            if |Zind| <= 3
                |
                |
                |       y - SD3pos
        Zind* = | 3 + ( ----------- )   if Zind > 3
                |         SD23pos
                |
                |
                |
                |        y - SD3neg
                | -3 + ( ----------- )  if Zind < -3
                |          SD23neg
        """

        numerator = (self.measurement / self.median)**self.skew - D(1.0)
        denominator = self.skew * self.coff
        z_score = numerator / denominator

        if D(z_score) > D(3):
            SD2pos = self.calc_stdev(2)
            SD3pos = self.calc_stdev(3)

            SD23pos = SD3pos - SD2pos

            z_score = 3 + ((self.measurement - SD3pos) / SD23pos)

            z_score = float(z_score.quantize(D('0.01')))

        elif D(z_score) < -3:
            SD2neg = self.calc_stdev(-2)
            SD3neg = self.calc_stdev(-3)

            SD23neg = SD2neg - SD3neg

            z_score = -3 + ((self.measurement - SD3neg) / SD23neg)
            z_score = float(z_score.quantize(D('0.01')))

        else:
            z_score = float(z_score.quantize(D('0.01')))

        return z_score



from azureml.core import Experiment, Run, Workspace
from config_deepensemble_1 import CONFIG

if __name__ == "__main__":

    workspace = Workspace.from_config()
    experiment = Experiment(workspace=workspace, name=CONFIG.EXPERIMENT_NAME)
    run = Run(experiment, CONFIG.RUN_ID)
    model = run.register_model(model_name=CONFIG.MODEL_NAME,
                               model_path='outputs')
    print('Model register successfully')



import argparse
import logging
import logging.config
import random
import shutil
from importlib import import_module
from pathlib import Path

import tensorflow as tf
from azureml.core.run import Run

from constants import DEFAULT_CONFIG, REPO_DIR

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d'))
logger.addHandler(handler)


def copy_dir(src: Path, tgt: Path, glob_pattern: str, should_touch_init: bool = False):
    logger.info("Creating temp folder")
    if tgt.exists():
        shutil.rmtree(tgt)
    tgt.mkdir(parents=True, exist_ok=True)
    if should_touch_init:
        (tgt / '__init__.py').touch(exist_ok=False)

    paths_to_copy = list(src.glob(glob_pattern))
    logger.info(f"Copying to {tgt} the following files: {str(paths_to_copy)}")
    for p in paths_to_copy:
        destpath = tgt / p.relative_to(src)
        destpath.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(p, destpath)


def is_offline_run(run: Run) -> bool:
    return run.id.startswith("OfflineRun")


# Get the current run.
RUN = Run.get_context()

from cgmml.common.evaluation.eval_utilities import (  # noqa: E402, F401
    is_ensemble_evaluation, is_multiartifact_evaluation)
from cgmml.common.evaluation.evaluation_classes import (  # noqa: E402, F401
    Evaluation, EnsembleEvaluation, MultiartifactEvaluation)
from cgmml.common.model_utils.run_initialization import OfflineRunInitializer, OnlineRunInitializer  # noqa: E402

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(levelname)s - %(message)s - %(pathname)s: line %(lineno)d')

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--qa_config_module", default=DEFAULT_CONFIG, help="Configuration file")
    args = parser.parse_args()
    qa_config_module = args.qa_config_module
    qa_config = import_module(qa_config_module)
else:
    qa_config_module = DEFAULT_CONFIG
    qa_config = import_module(qa_config_module)
logger.info('Using the following config: %s', qa_config_module)


MODEL_CONFIG = qa_config.MODEL_CONFIG
EVAL_CONFIG = qa_config.EVAL_CONFIG
DATA_CONFIG = qa_config.DATA_CONFIG
RESULT_CONFIG = qa_config.RESULT_CONFIG
FILTER_CONFIG = qa_config.FILTER_CONFIG if getattr(qa_config, 'FILTER_CONFIG', False) else None


if __name__ == "__main__":
    # Make experiment reproducible
    tf.random.set_seed(EVAL_CONFIG.SPLIT_SEED)
    random.seed(EVAL_CONFIG.SPLIT_SEED)

    if is_offline_run(RUN):
        OUTPUT_CSV_PATH = str(REPO_DIR / 'data' / RESULT_CONFIG.SAVE_PATH)
        initializer = OfflineRunInitializer(DATA_CONFIG, EVAL_CONFIG)
    else:
        OUTPUT_CSV_PATH = RESULT_CONFIG.SAVE_PATH
        initializer = OnlineRunInitializer(DATA_CONFIG, EVAL_CONFIG, RUN)

    if is_ensemble_evaluation(MODEL_CONFIG):
        MODEL_BASE_DIR = (REPO_DIR / 'data' / MODEL_CONFIG.EXPERIMENT_NAME) if is_offline_run(RUN) else Path('.')
        eval_class = EnsembleEvaluation
        descriptor = MODEL_CONFIG.EXPERIMENT_NAME
    else:
        MODEL_BASE_DIR = REPO_DIR / 'data' / MODEL_CONFIG.RUN_ID if is_offline_run(RUN) else Path('.')
        eval_class = MultiartifactEvaluation if is_multiartifact_evaluation(DATA_CONFIG) else Evaluation
        descriptor = MODEL_CONFIG.RUN_ID
    evaluation = eval_class(MODEL_CONFIG, DATA_CONFIG, MODEL_BASE_DIR, initializer.dataset_path)
    evaluation.get_the_model_path(initializer.workspace)

    # Get the QR-code paths
    qrcode_paths = evaluation.get_the_qr_code_path()
    if getattr(EVAL_CONFIG, 'DEBUG_RUN', False) and len(qrcode_paths) > EVAL_CONFIG.DEBUG_NUMBER_OF_SCAN:
        qrcode_paths = qrcode_paths[:EVAL_CONFIG.DEBUG_NUMBER_OF_SCAN]
        logger.info("Executing on %d qrcodes for FAST RUN", EVAL_CONFIG.DEBUG_NUMBER_OF_SCAN)

    dataset_evaluation, paths_belonging_to_predictions = evaluation.prepare_dataset(qrcode_paths, FILTER_CONFIG)
    prediction_array = evaluation.get_prediction_(evaluation.model_path_or_paths, dataset_evaluation)
    logger.info("Prediction made by model on the depthmaps...")
    logger.info(prediction_array)

    df = evaluation.prepare_dataframe(paths_belonging_to_predictions, prediction_array, RESULT_CONFIG)
    evaluation.evaluate(df, RESULT_CONFIG, EVAL_CONFIG, OUTPUT_CSV_PATH, descriptor)

    # Done.
    initializer.run.complete()
